"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import logging
from datetime import datetime

from rapid.lib.constants import StatusConstants
from rapid.lib.exceptions import QueueHandlerShouldSleep
from rapid.lib.framework.injectable import Injectable
from rapid.master.master_configuration import MasterConfiguration
from rapid.release.release_service import ReleaseService
from rapid.workflow.action_instances_service import ActionInstanceService
from rapid.workflow.queue_handlers.queue_handler import QueueHandler
from rapid.workflow.queue_service import QueueService

logger = logging.getLogger("rapid")


class Queue(Injectable):
    __injectables__ = {'queue_service': QueueService,
                       'action_instance_service': ActionInstanceService,
                       'rapid_config': MasterConfiguration,
                       'queue_constants': None,
                       'release_service': ReleaseService}

    def __init__(self, queue_service, action_instance_service, rapid_config, queue_constants, release_service):
        """
        :param queue_service:
        :type queue_service:
        :param action_instance_service:
        :type action_instance_service: rapid.workflow.action_instances_service.ActionInstanceService
        :type rapid_config: rapid.master.master_configuration.MasterConfiguration
        :type handler_constants: QueueHandlerConstants
        :type release_service: ReleaseService
        """
        self.queue_service = queue_service
        self.action_instance_service = action_instance_service
        self.rapid_config = rapid_config
        self.handler_constants = queue_constants
        self.release_service = release_service

    @property
    def queue_handlers(self): # type: () -> list[QueueHandler]
        return self.handler_constants.queue_handlers

    def process_queue(self, clients):
        sleeping_queue_handlers = []
        for work_request in self.queue_service.get_current_work():
            for queue_handler in self.queue_handlers:
                if queue_handler in sleeping_queue_handlers:
                    continue

                if queue_handler.can_process_work_request(work_request):
                    try:
                        queue_handler.process_work_request(work_request, clients)
                    except QueueHandlerShouldSleep:
                        sleeping_queue_handlers.append(queue_handler)
                    except Exception as exception:
                        logger.error(exception)
                        self.action_instance_service.edit_action_instance(work_request.action_instance_id, {'status_id': StatusConstants.FAILED,
                                                                                                            'start_date': datetime.utcnow(),
                                                                                                            'end_date': datetime.utcnow()})
                    break

    def verify_still_working(self, clients):
        for action_instance in self.queue_service.get_verify_working(self.rapid_config.queue_consider_late_time):
            for queue_handler in self.queue_handlers:
                if queue_handler.can_process_action_instance(action_instance):
                    try:
                        queue_handler.process_action_instance(action_instance, clients)
                    except Exception as exception:
                        logger.error(exception)
                        self.action_instance_service.edit_action_instance(action_instance['id'], {'status_id': StatusConstants.FAILED,
                                                                                                  'start_date': datetime.utcnow(),
                                                                                                  'end_date': datetime.utcnow()})
                    break

    def reconcile_pipeline_instances(self):
        self.action_instance_service.reconcile_pipeline_instances()

    def reconcile_releases(self):
        return self.release_service.reconcile_releases()

    def cancel_worker(self, action_instance):
        # type: (dict) -> bool
        for handler in self.queue_handlers:
            if handler.can_process_action_instance(action_instance):
                return handler.cancel_worker(action_instance)
        return False



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=bare-except,too-many-locals,unused-import,unused-variable,too-many-branches,too-many-statements,too-many-nested-blocks
from sqlalchemy import and_
from sqlalchemy.orm import joinedload, subqueryload
from sqlalchemy.sql.expression import select, union_all
from sqlalchemy.sql.functions import func
from sqlalchemy.sql.schema import Column

from rapid.lib.results_serializer import ResultsSerializer
from rapid.qa.data.models import QaProduct
from rapid.lib.utils import ORMUtil
from rapid.lib.constants import Constants, StatusConstants
from rapid.lib import get_db_session
from rapid.master.data.database.dal.general_dal import GeneralDal
from rapid.qa.data.models import QaTestHistory, QaStatusSummary, Stacktrace, QaTest, QaArea, QaFeature, QaBehaviorPoint, QaTestMapping
from rapid.workflow.data.models import Status
from rapid.ci.data.models import Vcs
from rapid.workflow.data.models import PipelineInstance


class QaDal(GeneralDal):
    def reset_results(self, action_instance_id, session):
        """

        :param action_instance_id:
        :type action_instance_id:
        :param session:
        :type session: sqlalchemy.orm.session.Session
        :return:
        :rtype:
        """
        session.execute(Stacktrace.__table__.delete().where(Stacktrace.qa_test_history_id.in_(
            session.query(QaTestHistory.id).filter(QaTestHistory.action_instance_id == action_instance_id))))

        session.execute(QaTestHistory.__table__.delete().where(QaTestHistory.action_instance_id == action_instance_id))
        session.execute(QaStatusSummary.__table__.delete().where(QaStatusSummary.action_instance_id == action_instance_id))

    def save_results(self, action_instance, session, post_data):
        return self._save_results(action_instance, session, post_data)

    def get_test_results(self, action_instance_id, filters=None, in_session=None):
        results = {}
        if in_session is None:
            for session in get_db_session():
                query = session.query(QaTestHistory).options(
                    joinedload(QaTestHistory.test)).options(
                        joinedload(QaTestHistory.status)).options(
                            joinedload(QaTestHistory.stacktrace)) \
                    .filter(QaTestHistory.action_instance_id == action_instance_id)
                query = ORMUtil.get_filtered_query(query, filters, QaTestHistory)

                for test_history in query.all():
                    status = test_history.status.name
                    results_array = []
                    if status not in results:
                        results[status] = results_array
                    else:
                        results_array = results[status]
                    results_array.append(test_history.serialize({QaTestHistory.__tablename__: ['test', 'status', 'stacktrace']}))
                return results
        else:
            query = in_session.query(QaTestHistory).options(
                joinedload(QaTestHistory.test)).options(
                    joinedload(QaTestHistory.status)).options(
                        joinedload(QaTestHistory.stacktrace)) \
                .filter(QaTestHistory.action_instance_id == action_instance_id)
            query = ORMUtil.get_filtered_query(query, filters, QaTestHistory)

            for test_history in query.all():
                status = test_history.status.name
                results_array = []
                if status not in results:
                    results[status] = results_array
                else:
                    results_array = results[status]
                results_array.append(test_history.serialize(['test', 'status']))
            return results
        return results

    def analyze_tests(self, pipeline_instance_id, json):
        """
        Analyze test results
        :param pipeline_instance_id:
        :type pipeline_instance_id: int
        :param json:
        :type json: dict
        :return:
        :rtype: dict
        """
        if json:
            # Deal with Areas
            # Deal with Features
            # Deal with Behavior Points
            for session in get_db_session():
                product = session.query(QaProduct) \
                    .join(Vcs) \
                    .join(PipelineInstance, PipelineInstance.pipeline_id == Vcs.pipeline_id) \
                    .filter(PipelineInstance.id == pipeline_instance_id).first()

                if product:
                    area_mapper = self._get_area_mapper(session, product.id, json)
                    return area_mapper

        return {}

    def _get_feature_mapper(self, session, json):
        feature_mapper = {}
        feature_keys = list()

        for area, features in json.items():
            feature_keys.extend(list(features.keys()))

        for feature in session.query(QaFeature).filter(QaFeature.in_(feature_keys)):
            feature_mapper[feature.name] = feature
            del feature_keys[feature.name]

        for feature_name in feature_keys:
            qa_feature = QaFeature(name=feature_name)
            session.add(qa_feature)
            feature_mapper[qa_feature.name] = qa_feature

        session.flush()
        return feature_mapper

    def _get_area_mapper(self, session, product_id, json):
        area_mapper = {}
        feature_mapper = {}
        bp_mapper = {}
        joiner_mapper = {}
        area_keys = []
        feature_keys = []
        bp_keys = []
        test_keys = []
        test_mapper = {}

        for area, area_map in json.items():
            area_keys.append(area)
            area_mapper[area] = {'result': area_map, 'model': None}
            for feature, feature_map in area_map.items():
                feature_keys.append(feature)
                feature_mapper[feature] = {'result': feature_map, 'model': None}
                for _bp, tests in feature_map.items():
                    bp_keys.append(_bp)
                    bp_mapper[_bp] = {'result': tests, 'model': None}
                    for test in tests:
                        test_keys.append(test['name'])
                        test['__key__'] = "{}:{}:{}".format(area, feature, _bp)
                        test['model'] = None
                        test_mapper[test['name']] = test

        if area_keys:
            for area in session.query(QaArea).filter(QaArea.name.in_(area_keys))\
                                             .filter(QaArea.product_id == product_id).all():
                try:
                    area_keys.remove(area.name)
                    area_mapper[area.name]['model'] = area
                except:
                    pass

        for area in area_keys:
            area_model = QaArea(name=area, product_id=product_id)
            session.add(area_model)
            area_mapper[area]['model'] = area_model

        session.flush()

        if feature_keys:
            for feature in session.query(QaFeature).filter(QaFeature.name.in_(feature_keys)).all():
                try:
                    feature_keys.remove(feature.name)
                except:
                    pass
                try:
                    feature_mapper[feature.name]['model'] = feature
                except:
                    pass

        for feature in feature_keys:
            feature_model = QaFeature(name=feature, product_id=product_id)
            session.add(feature_model)
            try:
                feature_mapper[feature]['model'] = feature_model
            except:
                pass

        session.flush()

        if bp_keys:
            for _bp in session.query(QaBehaviorPoint).filter(QaBehaviorPoint.name.in_(bp_keys)).all():
                try:
                    bp_keys.remove(_bp.name)
                except:
                    pass
                try:
                    bp_mapper[_bp.name]['model'] = _bp
                except:
                    pass

        for _bp in bp_keys:
            bp_model = QaBehaviorPoint(name=_bp, product_id=product_id)
            session.add(bp_model)
            try:
                bp_mapper[_bp]['model'] = bp_model
            except:
                pass

        session.flush()

        if test_keys:
            for test, qa_test_mapping in session.query(QaTest, QaTestMapping) \
                    .outerjoin(QaTestMapping, QaTest.id == QaTestMapping.test_id) \
                    .filter(QaTest.name.in_(test_keys)).all():
                if test.name in test_mapper:
                    (area, feature, _bp) = test_mapper[test.name]['__key__'].split(':', 2)
                    if qa_test_mapping is None:
                        qa_test_mapping = QaTestMapping(area_id=area_mapper[area]['model'].id,
                                                        test_id=test.id,
                                                        feature_id=feature_mapper[feature]['model'].id,
                                                        behavior_id=bp_mapper[_bp]['model'].id)
                        session.add(qa_test_mapping)
                    else:
                        if qa_test_mapping.area_id != area_mapper[area]['model'].id:
                            qa_test_mapping.area_id = area_mapper[area]['model'].id
                        if qa_test_mapping.feature_id != feature_mapper[feature]['model'].id:
                            qa_test_mapping.feature_id = feature_mapper[feature]['model'].id
                        if qa_test_mapping.behavior_id != bp_mapper[_bp]['model'].id:
                            qa_test_mapping.behavior_id = bp_mapper[_bp]['model'].id

        session.flush()
        session.commit()

        return {"message": "Success!"}

    def _get_summary_results(self, results):
        try:
            return results[Constants.RESULTS_SUMMARY]
        except:
            pass
        return None

    def _save_summary(self, action_instance, results, session):
        try:
            if Constants.RESULTS_SUMMARY in results:
                summary = results[Constants.RESULTS_SUMMARY]

                if hasattr(summary, 'keys'):
                    statuses = {}
                    for status in session.query(Status).filter(Status.name.in_(list(summary.keys()))):
                        statuses[status.name] = status

                    for (key, value) in summary.items():
                        if key in statuses:
                            session.add(QaStatusSummary(status_id=statuses[key].id, action_instance_id=action_instance.id, count=value))
                    session.flush()
                del results[Constants.RESULTS_SUMMARY]
        except:
            import traceback
            traceback.print_exc()
        return results

    def _save_results(self, action_instance, session, post_data):
        if 'results' in post_data:
            try:
                failures_count = False
                try:
                    failures_count = self._get_summary_results(post_data['results'])[Constants.FAILURES_COUNT]
                except:
                    pass

                results = self._save_summary(action_instance, post_data['results'], session)

                should_create = dict(results)
                test_cache = {}
                for qa_test in session.query(QaTest).filter(QaTest.name.in_(list(results.keys()))):
                    if qa_test.name in should_create:
                        del should_create[qa_test.name]
                        test_cache[qa_test.name] = qa_test

                for name in should_create.keys():
                    qa_test = QaTest(name=name)
                    session.add(qa_test)
                    test_cache[qa_test.name] = qa_test

                session.flush()

                status_cache = {}
                for test, value in results.items():
                    status_id = None
                    if 'status' not in value:
                        status_id = StatusConstants.UNKNOWN

                    if 'status' in value:
                        if value['status'] not in status_cache:
                            status = session.query(Status).filter(func.lower(Status.name) == func.lower(value['status'])).first()
                            if status:
                                status_id = status.id
                                status_cache[status.name] = status
                            else:
                                status_id = StatusConstants.UNKNOWN
                        else:
                            status_id = status_cache[value['status']].id

                    if not failures_count or status_id == StatusConstants.FAILED:
                        qa_test_history = QaTestHistory(test_id=test_cache[test].id,
                                                        pipeline_instance_id=action_instance.pipeline_instance_id,
                                                        action_instance_id=action_instance.id,
                                                        status_id=status_id,
                                                        duration=value['time'] if 'time' in value and value['time'] else 0)

                        session.add(qa_test_history)

                        session.flush()

                        if 'stacktrace' in value:
                            session.add(Stacktrace(qa_test_history_id=qa_test_history.id, stacktrace=value['stacktrace']))

                session.commit()
            except:
                import traceback
                traceback.print_exc()

    def get_qa_testmap_coverage(self, pipeline_instance_id):
        objects = []
        for session in get_db_session():
            query = session.query(QaTestMapping) \
                .options(joinedload(QaTestMapping.feature, innerjoin=True)) \
                .options(joinedload(QaTestMapping.behavior_point, innerjoin=True)) \
                .options(joinedload(QaTestMapping.test, innerjoin=True)) \
                .options(joinedload(QaTestMapping.area, innerjoin=True)) \
                .join(PipelineInstance, PipelineInstance.id == pipeline_instance_id) \
                .join(Vcs, Vcs.pipeline_id == PipelineInstance.pipeline_id) \
                .join(QaProduct, QaProduct.vcs_id == Vcs.id) \
                .join(QaArea, and_(QaArea.product_id == QaProduct.id, QaArea.id == QaTestMapping.area_id))
            objects = ResultsSerializer.serialize_results(query.all(), {QaTestMapping.__tablename__: ['area',
                                                                                                      'feature',
                                                                                                      'behavior_point',
                                                                                                      'test']})
        return objects



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""


class Communicator(object):

    def __init__(self, server_uri, flask_app=None):
        self.server_uri = server_uri
        self.flask_app = flask_app

    def get_uri(self, uri_stem):
        return self.server_uri + uri_stem

    def api_uri(self, uri_stem):
        return self.get_uri("/api/{}".format(uri_stem))



"""qa test mapping

Revision ID: 593023a41b6d
Revises: e2a50a08334a
Create Date: 2016-05-19 16:09:12.356379

"""

# revision identifiers, used by Alembic.

revision = '593023a41b6d'
down_revision = '63ed28aeec3d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('qa_areas',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=250), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_qa_areas_id'), 'qa_areas', ['id'], unique=False)
    op.create_index(op.f('ix_qa_areas_name'), 'qa_areas', ['name'], unique=False)
    op.create_table('qa_behavior_points',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=400), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_qa_behavior_points_id'), 'qa_behavior_points', ['id'], unique=False)
    op.create_index(op.f('ix_qa_behavior_points_name'), 'qa_behavior_points', ['name'], unique=False)
    op.create_table('qa_features',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=250), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_qa_features_id'), 'qa_features', ['id'], unique=False)
    op.create_index(op.f('ix_qa_features_name'), 'qa_features', ['name'], unique=False)
    op.create_table('tags',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_tags_id'), 'tags', ['id'], unique=False)
    op.create_index(op.f('ix_tags_name'), 'tags', ['name'], unique=False)
    op.create_table('qa_test_mappings',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('area_id', sa.Integer(), nullable=False),
    sa.Column('feature_id', sa.Integer(), nullable=False),
    sa.Column('behavior_id', sa.Integer(), nullable=False),
    sa.Column('test_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['area_id'], ['qa_areas.id'], ),
    sa.ForeignKeyConstraint(['behavior_id'], ['qa_behavior_points.id'], ),
    sa.ForeignKeyConstraint(['feature_id'], ['qa_features.id'], ),
    sa.ForeignKeyConstraint(['test_id'], ['qa_tests.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('area_id', 'feature_id', 'behavior_id', 'test_id', name='uix_qatm')
    )
    op.create_index(op.f('ix_qa_test_mappings_id'), 'qa_test_mappings', ['id'], unique=False)
    op.create_table('qa_test_map_tags',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('qa_testmap_id', sa.Integer(), nullable=False),
    sa.Column('tag_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['qa_testmap_id'], ['qa_test_mappings.id'], ),
    sa.ForeignKeyConstraint(['tag_id'], ['tags.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_qa_test_map_tags_id'), 'qa_test_map_tags', ['id'], unique=False)
    op.create_index(op.f('ix_qa_test_map_tags_qa_testmap_id'), 'qa_test_map_tags', ['qa_testmap_id'], unique=False)
    op.create_index(op.f('ix_qa_test_map_tags_tag_id'), 'qa_test_map_tags', ['tag_id'], unique=False)
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('qa_test_map_tags')
    op.drop_index(op.f('ix_qa_test_mappings_id'), table_name='qa_test_mappings')
    op.drop_table('qa_test_mappings')
    op.drop_index(op.f('ix_tags_name'), table_name='tags')
    op.drop_index(op.f('ix_tags_id'), table_name='tags')
    op.drop_table('tags')
    op.drop_index(op.f('ix_qa_features_name'), table_name='qa_features')
    op.drop_index(op.f('ix_qa_features_id'), table_name='qa_features')
    op.drop_table('qa_features')
    op.drop_index(op.f('ix_qa_behavior_points_name'), table_name='qa_behavior_points')
    op.drop_index(op.f('ix_qa_behavior_points_id'), table_name='qa_behavior_points')
    op.drop_table('qa_behavior_points')
    op.drop_index(op.f('ix_qa_areas_name'), table_name='qa_areas')
    op.drop_index(op.f('ix_qa_areas_id'), table_name='qa_areas')
    op.drop_table('qa_areas')

    ### end Alembic commands ###



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import logging

from sqlalchemy import and_, exists

from rapid.ci.data.models import Commit
from rapid.lib.constants import ModuleConstants, StatusConstants, status_type_severity_mapping
from rapid.lib.framework.injectable import Injectable
from rapid.lib import get_db_session
from rapid.release.data.models import Step, Release

logger = logging.getLogger("rapid")


class ReleaseDal(Injectable):
    __injectables__ = {ModuleConstants.WORKFLOW_MODULE: None, ModuleConstants.CI_MODULE: None}

    def __init__(self, workflow_module, ci_module):
        """

        :type workflow_module: :class:`rapid.lib.modules.modules.WorkflowModule`
        :type ci_module: :class:`rapid.lib.modules.modules.CiModule`
        :return:
        """
        self.workflow_module = workflow_module
        self.ci_module = ci_module

    def get_release_by_id(self, release_id, in_session=None, with_for_update=False):
        if in_session is None:
            for session in get_db_session():
                if with_for_update:
                    return session.query(Release).with_for_update().filter(Release.id == release_id).first().serialize()
                else:
                    return session.query(Release).with_for_update().filter(Release.id == release_id).first().serialize()
        else:
            if with_for_update:
                return in_session.query(Release).with_for_update().filter(Release.id == release_id).first()
            else:
                return in_session.query(Release).with_for_update().filter(Release.id == release_id).first()

    def set_release_step(self, release_id, step_id, status, session=None):
        assert release_id is not None
        assert step_id is not None
        assert status is not None

        if session is None:
            for db_session in get_db_session():
                release = self.get_release_by_id(release_id, db_session, with_for_update=True)
                status_entry = self.workflow_module.get_status_by_name(status, db_session)

                if status_entry and status_entry.id:
                    found_step = self._set_internal_release_step(release_id, step_id, status, db_session)

                    self._finish_release_if_necessary(release, step_id, status_entry)

                    db_session.commit()
                    if found_step is not None:
                        return found_step.serialize()
        else:
            status_entry = self.workflow_module.get_status_by_name(status, session)
            if status_entry and status_entry.id:
                release = self.get_release_by_id(release_id, session)
                step = self._set_internal_release_step(release_id, step_id, status, session).serialize()

                self._finish_release_if_necessary(release, step_id, status_entry)
                return step
        return None

    def _finish_release_if_necessary(self, release, step_id, status_entry):
        count_complete = 0
        found_step = None
        highest_status = None
        for step in release.steps:
            if step.custom_id == step_id:
                step.status_id = status_entry.id
                found_step = step
            if step.status_id >= StatusConstants.SUCCESS:
                count_complete += 1
                if step.status.type in status_type_severity_mapping:
                    if highest_status is None or status_type_severity_mapping[highest_status.type] > status_type_severity_mapping[step.status.type]:
                        highest_status = step.status

        if count_complete == len(release.steps):
            release.status_id = highest_status.id if highest_status is not None else StatusConstants.SUCCESS
        elif release.status_id < StatusConstants.SUCCESS:
            release.status_id = StatusConstants.INPROGRESS
        return found_step

    def _set_internal_release_step(self, release_id, step_id, status, session):
        status_entry = self.workflow_module.get_status_by_name(status, session)
        if status_entry and status_entry.id:
            step = self.get_step_by_release_and_step_id(release_id, step_id, session)
            if step and step.id:
                step.status_id = status_entry.id
                return step
        raise Exception("Something didn't work!")

    def get_step_by_release_and_step_id(self, release_id, step_id, in_session=None):
        if in_session is not None:
            return in_session.query(Step).filter(Step.release_id == release_id)\
                .filter(Step.custom_id == step_id).first()
        for session in get_db_session():
            return session.query(Step).filter(Step.release_id == release_id) \
                .filter(Step.custom_id == step_id).first().serialize()

    def get_release_by_commit_identifier(self, commit_identifier, in_session=None):
        if in_session is None:
            for session in get_db_session():
                release = session.query(Release).filter(Release.commit_id == Commit.id)\
                    .filter(Commit.commit_identifier == commit_identifier).first()
                return release.serialize() if release is not None else None
        else:
            return in_session.query(Release).filter(Release.commit_id == commit_identifier).first()

    def mark_step_for_commit(self, commit_identifier, step_custom_id, status_name):
        commit = self.ci_module.get_by_identifier(commit_identifier)
        if commit is not None:
            for session in get_db_session():
                release = self.get_release_by_commit_identifier(commit_identifier)
                if release is not None:
                    step = self.set_release_step(release.id, step_custom_id, status_name, session)
                    session.commit()
                    return step

        logger.debug("Step is not the head of the release.")
        return None

    def reconcile_releases(self):
        releases = []
        for session in get_db_session():
            for release in session.query(Release).join(Step).filter(and_(Release.status_id < StatusConstants.SUCCESS,
                                                                         ~exists().where(Step.status_id < StatusConstants.SUCCESS))).all():
                release.status_id = StatusConstants.SUCCESS
                releases.append(release.serialize())
            session.commit()
        return releases



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import os


class Features(object):
    TESTING = "TESTING"
    GITHUB_ALERTS = "GITHUB_ALERTS"

    full_features = [TESTING, GITHUB_ALERTS]

    enabled_features = []

    @staticmethod
    def get_enabled_features():
        features_enabled = []
        for feature in Features.full_features:
            if Features.is_enabled(feature):
                features_enabled.append(feature)
        return features_enabled

    @staticmethod
    def is_enabled(key):
        if hasattr(Features, key):
            if key in Features.enabled_features:
                return True
            if key in os.environ:
                return os.environ[key].lower().strip() == "true"
        return False



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import re

try:
    import simplejson as json
except ImportError:
    import json


class WorkRequest(object):
    def __init__(self, data=None):
        self._grain = None
        self.cmd = None
        self.executable = None
        self.args = None
        self.environment = {}
        self.headers = {}
        self.action_instance_id = None
        self.pipeline_instance_id = None
        self.workflow_instance_id = None
        self.slice = None
        self.configuration = None

        if data:
            self.prepare_to_send(data)

    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def generate_from_request(self, request):
        for (attr, value) in request.json.items():
            try:
                tmp = attr
                if tmp == 'grain':
                    tmp = '_grain'
                getattr(self, tmp)
                setattr(self, tmp, value)
            except (TypeError, AttributeError):
                pass

    def prepare_to_send(self, data_in):
        if isinstance(data_in, dict):
            for attr in vars(self).keys():
                try:
                    if attr == '_grain':
                        for g_attr in ['_grain', 'grain']:
                            if g_attr in data_in:
                                setattr(self, '_grain', data_in[g_attr])
                    else:
                        setattr(self, attr, data_in[attr])
                except (TypeError, AttributeError, KeyError):
                    pass
        else:
            for attr in vars(self).keys():
                try:
                    if attr == '_grain':
                        for g_attr in ['_grain', 'grain']:
                            if hasattr(data_in, g_attr):
                                setattr(self, attr, getattr(data_in, g_attr))
                    else:
                        setattr(self, attr, getattr(data_in, attr))
                except (TypeError, AttributeError, KeyError):
                    pass

        self.set_headers()

    def set_headers(self):
        self.headers['content-type'] = 'application/json'

    @property
    def grain(self):
        if '{' in self._grain:  # pylint: disable=unsupported-membership-test
            match = re.findall(r"\{([\w\d]*)\}", self._grain)
            if match:
                tmp = self._grain
                for param in match:
                    if param in self.environment:
                        tmp = tmp.replace('{{{}}}'.format(param), self.environment[param])
                return tmp
        return self._grain


class WorkRequestEncoder(json.JSONEncoder):
    def default(self, o):  # pylint: disable=method-hidden
        state = o.__dict__.copy()
        state['grain'] = state.pop('_grain', None)
        return state



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import os
import uuid
import tempfile

from requests.auth import HTTPBasicAuth

from ..lib.configuration import Configuration

# pylint: disable=too-many-instance-attributes, too-few-public-methods


class ClientConfiguration(Configuration):

    def __init__(self, file_name=None):
        self.port = None
        self.workspace = None
        self.master_uri = None
        self.registration_rate = None
        self.executor_count = None
        self.grains = None
        self.grain_restrict = None
        self.quarantine_directory = None
        self.register_api_key = None
        self.api_key = None
        self.install_uri = None
        self.install_options = None
        self.get_files_basic_auth = None
        self.use_ssl = None
        self.verify_certs = None
        self.log_file = None
        self.os_path_override = None
        self.log_to_directory = False

        self.is_single_use = False

        super(ClientConfiguration, self).__init__(file_name)
        self.executors = []

    @property
    def section_mapping(self):
        return {
            'client': {
                'port': [8081, int],
                'workspace': [os.path.join(tempfile.gettempdir(), 'rapid', 'workspace')],
                'master_uri': ['http://rapidci.local'],
                'registration_rate': [180, int],
                'executor_count': [2, int],
                'os_path_override': [None],
                'grains': [None],
                'grain_restrict': [False, bool],
                'quarantine_directory': [os.path.join(tempfile.gettempdir(), 'rapid', 'quarantine')],
                'register_api_key': [None],
                'api_key': [str(uuid.uuid3(uuid.NAMESPACE_OID, 'ClientApiKey!')).replace('-', '')],
                'install_uri': ['https://pypi.python.org/pypi/'],
                'install_options': [''],
                'get_files_basic_auth': [None, list, ':'],
                'log_to_directory': [None, str]
            },
            'general': {
                'use_ssl': [False, bool],
                'verify_certs': [True, bool],
                'log_file': [None],
                'extensions': [None, list, ',']
            }
        }

    def _set_values(self, parser):
        super(ClientConfiguration, self)._set_values(parser)

        if self.get_files_basic_auth:
            try:
                self.get_files_basic_auth = HTTPBasicAuth(self.get_files_basic_auth[0], self.get_files_basic_auth[1])
            except Exception:  # pylint: disable=broad-except
                pass

        if self.os_path_override:
            try:
                os.environ.update({'os_path_override':self.os_path_override})
            except Exception:  # pylint: disable=broad-except
                pass





import subprocess

import os

import yaml
from yaml.scanner import ScannerError

from rapid.lib.constants import StatusConstants
from rapid.lib.features import Features
from rapid.lib.framework.injectable import Injectable
from rapid.lib.work_request import WorkRequest
from rapid.master.master_configuration import MasterConfiguration
from rapid.workflow.action_instances_service import ActionInstanceService
from rapid.workflow.queue_handlers.container_handlers.container_handler import ContainerHandler
from rapid.workflow.queue_handlers.container_handlers.docker_configuration import DockerConfiguration
from rapid.workflow.queue_handlers.queue_handler_register import register_queue_handler


@register_queue_handler
class DockerQueueHandler(ContainerHandler, Injectable):
    __injectables__ = {'rapid_config': None, 'action_instance_service': ActionInstanceService}
    
    @property
    def container_identifier(self):
        return 'docker'

    def __init__(self, rapid_config, action_instance_service):
        # type: (MasterConfiguration, ActionInstanceService) -> None
        super(DockerQueueHandler, self).__init__(rapid_config)
        self.action_instance_service = action_instance_service

    def process_action_instance(self, action_instance, clients):
        # TODO - Still don't know how to verify this is working.
        pass

    def process_work_request(self, work_request, clients):
        # type: (WorkRequest, list) -> bool
        stdout = subprocess.PIPE
        stderr = subprocess.STDOUT

        (grain_type, image) = self._get_grain_type_split(work_request.grain)

        command = "docker run {} {}".format(
            self._get_docker_parameters(work_request),
            image
        )
        assignment = {
            'assigned_to': 'docker://',
            'status_id': StatusConstants.INPROGRESS
        }
        try:
            child_process = subprocess.Popen(command.split(' '), stderr=stderr, stdout=stdout)
            if child_process.poll():
                raise OSError('Something went wrong')
        except OSError:
            assignment['status_id'] = StatusConstants.FAILED
        self.action_instance_service.edit_action_instance(work_request.action_instance_id, assignment)
        return assignment['status_id'] == StatusConstants.INPROGRESS

    def _get_docker_parameters(self, work_request):
        try:
            docker_configuration = DockerConfiguration(**yaml.safe_load(work_request.configuration))
            return docker_configuration.get_string_rep(work_request)
        except (TypeError, AttributeError, ScannerError):
            pass

        return ''

    def get_environment(self, work_request):
        environment = {}
        for key in os.environ:
            environment[key] = os.environ[key]

        if work_request.environment:
            for key, value in work_request.environment.items():
                try:
                    environment[key.encode('ascii', 'ignore')] = value.encode('ascii', 'ignore')
                except AttributeError:
                    pass
        environment['PYTHONUNBUFFERED'] = "true"
        environment['pipeline_instance_id'] = str(work_request.pipeline_instance_id)
        environment['action_instance_id'] = str(work_request.action_instance_id)
        environment['workflow_instance_id'] = str(work_request.workflow_instance_id)
        environment['slice'] = str(work_request.slice)
        environment['RAPID_FEATURES'] = ",".join(Features.get_enabled_features())
        environment['WORKSPACE'] = self.rapid_config.workspace

        return environment

    def cancel_worker(self, action_instance):  # type: (dict) -> bool
        return True



import re

from rapid.lib.work_request import WorkRequest


class DockerConfiguration(object):
    _acceptable = [
        'add-host', 'env', 'volume'
    ]

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if key in self._acceptable:
                setattr(self, key, value)

    def get_string_rep(self, work_request):
        # type: (WorkRequest) -> str
        string_list = []
        keys = sorted(self.__dict__.keys())
        for key in keys:
            value = getattr(self, key)
            if isinstance(value, list):
                string_list.extend(['--{}={}'.format(key, self._get_substituted_value(work_request, tmp_value)) for tmp_value in value])
            else:
                string_list.append('--{}={}'.format(key, self._get_substituted_value(work_request, value)))
        return ' '.join(string_list)

    def _get_substituted_value(self, work_request, string):
        if '(' in string:
            try:
                for match in re.findall(r'\([\w_]*\)', string):
                    token = match.split('(')[1].split(')')[0]
                    try:
                        string = string.replace(match, '{}={}'.format(token, getattr(work_request, token)))
                    except AttributeError:
                        pass
            except (TypeError, AttributeError) as exception:
                pass
        return string



import json
from rapid.lib.configuration import Configuration, ConfigurationDefaults


class ECSConfiguration(Configuration):
    NOT_SET = "__NOT_SET__"

    def __init__(self, file_name=None):
        self.maximum_task_time = 20
        self.aws_access_key_id = None
        self.aws_secret_access_key = None
        self.aws_session_token = None
        self.region_name = None

        self.cluster = None
        self.taskDefinition = None
        self.count = None
        self.startedBy = None
        self.group = None
        self.placementConstraints = None
        self.placementStrategy = None
        self.launchType = None
        self.platformVersion = None
        self.networkConfiguration = None
        self.tags = None
        self.enableECSManagedTags = None
        self.propogateTags = None

        super(ECSConfiguration, self).__init__(file_name)

    @property
    def section_mapping(self):
        default_config = ConfigurationDefaults(self.NOT_SET, str)
        return {
            'general': {
                'maximum_task_time': ConfigurationDefaults(20*60, int)  # 20 minutes
            },
            'aws_credentials': {
                'aws_access_key_id': default_config,
                'aws_secret_access_key': default_config,
                'aws_session_token': default_config,
                'region_name': default_config,
            },
            'task_defaults': {
                'cluster': default_config,
                'taskDefinition': default_config,
                'overrides': ConfigurationDefaults(self.NOT_SET, dict),
                'count': default_config,
                'startedBy': default_config,
                'group': default_config,
                'placementConstraints': ConfigurationDefaults(self.NOT_SET, list),
                'placementStrategy': ConfigurationDefaults(self.NOT_SET, list),
                'launchType': default_config,
                'platformVersion': default_config,
                'networkConfiguration': ConfigurationDefaults(self.NOT_SET, dict),
                'tags': ConfigurationDefaults(self.NOT_SET, list),
                'enableECSManagedTags': ConfigurationDefaults(self.NOT_SET, bool),
                'propogateTags': default_config
            }
        }

    @property
    def aws_credentials(self):
        # type: () -> dict
        return self._get_section_values('aws_credentials')

    @property
    def default_task_definition(self):
        # type: () -> dict
        return self._get_section_values('task_defaults')

    def _get_section_values(self, section):
        return {key: getattr(self, key) for key in list(self.section_mapping[section].keys()) if getattr(self, key) != self.NOT_SET}

    def _handle_normal_value(self, parser, key, section, type_cast):
        new_value = parser.get(section, key)
        try:
            if type_cast in [list, dict]:
                setattr(self, key, type_cast(json.loads(new_value)))
            else:
                setattr(self, key, type_cast(new_value))
        except ValueError:
            setattr(self, key, type_cast(new_value))



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from xml.dom.minidom import parseString
from rapid.lib.constants import Constants

from .abstract_parser import AbstractParser


class XUnitParser(AbstractParser):
    @staticmethod
    def get_type():
        return "XUnit"

    def _parse_lines(self, lines):  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        """
        Removes the first line, being the identifier, then parses the rest of the lines
        :param lines: array of the lines of the file.
        :return: dict of results.
        """
        results = {}
        dom = parseString("".join(lines))
        summary = self.prepare_summary()

        for testsuite in dom.getElementsByTagName('testsuite'):
            for node in testsuite.childNodes:
                if node.nodeName != 'testcase':
                    continue
                testcase = node
                name = ''
                try:
                    class_name = ''
                    for class_attr in ['classname', 'class']:
                        class_name = testcase.getAttribute(class_attr)
                        if class_name:
                            break

                    name = "{}~{}".format(class_name, testcase.getAttribute('name'))
                except AttributeError:
                    import traceback
                    traceback.print_exc()

                if self.workspace:
                    name = name.replace(self.workspace, '')

                test_result = {
                    'status': Constants.STATUS_SUCCESS,
                    'time': testcase.getAttribute('time')
                }

                failure_tags = testcase.getElementsByTagName('failure')
                skipped_tags = testcase.getElementsByTagName('skipped')
                error_tags = testcase.getElementsByTagName('error')

                if self.failures_only and (not failure_tags and not error_tags):
                    continue

                if failure_tags:
                    test_result['status'] = Constants.STATUS_FAILED
                    test_result['stacktrace'] = ""

                    count = 0
                    length = len(failure_tags)
                    for failure in failure_tags:
                        test_result['stacktrace'] += "\n".join(
                            [failure.firstChild.nodeValue if failure.firstChild is not None and hasattr(failure.firstChild, 'nodeValue')
                             else failure.attributes['message'].value])
                        count += 1
                        if length > count:
                            test_result['stacktrace'] += "\n"

                    summary[Constants.STATUS_FAILED] += 1
                elif error_tags:
                    test_result['status'] = Constants.STATUS_FAILED
                    test_result['stacktrace'] = ""

                    count = 0
                    length = len(failure_tags)
                    for failure in error_tags:
                        test_result['stacktrace'] += "\n".join(
                            ["Error - " + (failure.firstChild.nodeValue if failure.firstChild is not None and hasattr(failure.firstChild, 'nodeValue')
                                           else failure.attributes['message'].value)])
                        count += 1
                        if length > count:
                            test_result['stacktrace'] += "\n"

                    summary[Constants.STATUS_FAILED] += 1
                elif skipped_tags:
                    test_result['status'] = Constants.STATUS_SKIPPED
                    summary[Constants.STATUS_SKIPPED] += 1
                else:
                    summary[Constants.STATUS_SUCCESS] += 1

                results[name] = test_result

        results['__summary__'] = summary

        return results



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
# pylint: disable=broad-except
import re
from abc import abstractmethod, ABCMeta

import pkgutil
from simpleeval import simple_eval


class EventHandlerFactory(object):

    @staticmethod
    def get_event_handler(in_type):
        """
        :param in_type:
        :type in_type: lib.Constants.EventTypes
        :return:
        :rtype: EventHandler
        """
        import rapid.workflow.events.handlers as handlers
        for importer, modname, ispkg in pkgutil.iter_modules(handlers.__path__):  # pylint: disable=unused-variable
            try:
                name = "{}.{}".format(handlers.__name__, modname)
                module = __import__(name, fromlist=[modname])
                class_name = ''.join([x.title() for x in modname.split('_')])
                tmp = getattr(module, class_name)
                if in_type in [tmp.get_event_type().value, tmp.get_event_type()]:
                    return tmp()
            except Exception:
                pass
        return None


class EventHandler(object):
    __metaclass__ = ABCMeta

    @staticmethod
    @abstractmethod
    def get_event_type():
        yield

    @abstractmethod
    def handle_event(self, pipeline_instance, action_instance, event):
        """
        :param pipeline_instance:
        :type pipeline_instance: rapid.workflow.data.models.PipelineInstance
        :param action_instance:
        :type action_instance: rapid.workflow.data.models.ActionInstance
        :param event:
        :type event: rapid.workflow.data.models.PipelineEvent
        :return:
        :rtype:
        """
        yield

    def passes_conditional(self, pipeline_instance, action_instance, conditional):
        """
        :param pipeline_instance:
        :type pipeline_instance: workflow.data.models.PipelineInstance
        :param conditional:
        :type conditional: str
        :return:
        :rtype:
        """
        if conditional is None or conditional == '':
            return True

        new_conditional = self._prepare_conditional(conditional, pipeline_instance, action_instance)
        return self._evaluate_condition(new_conditional)

    def _prepare_conditional(self, conditional, pipeline_instance, action_instance):
        """
        :param conditional:
        :type conditional: str
        :param parameters:
        :type parameters: dict
        :return:
        :rtype:
        """
        new_conditional = conditional
        for parameter, value in pipeline_instance.get_parameters_dict().items():
            if "{{{}}}".format(parameter) in new_conditional:
                new_conditional = new_conditional.replace("{{{}}}".format(parameter), str(value))

        if 'pipelineInstance.' in new_conditional:
            new_conditional = self._translate_string_for_pipeline_instance(pipeline_instance, new_conditional)

        if 'actionInstance.' in new_conditional:
            new_conditional = self._translate_string_for_action_instance(action_instance, new_conditional)

        return new_conditional

    def _evaluate_condition(self, condition):
        try:
            return simple_eval(condition)
        except Exception:
            return False

    def _translate_string_for_pipeline_instance(self, pipeline_instance, string):
        return self._translate_string_for_key_string('pipelineInstance', pipeline_instance, string)

    def _translate_string_for_action_instance(self, action_instance, string):
        return self._translate_string_for_key_string('actionInstance', action_instance, string)

    def _translate_string_for_key_string(self, key_string, obj_instance, string):
        new_string = string
        try:
            for term in re.findall(r'(' + key_string + r'\.[^\s^}]{0,})', new_string):
                replaced_term = term.replace('{}.'.format(key_string), '', 1)
                new_string = new_string.replace(term, str(self._get_attribute_trait(obj_instance, replaced_term)))
        except Exception:
            pass
        return new_string

    def _get_attribute_trait(self, obj, trait):
        if '.' in trait:
            sp_obj = trait.split('.', 2)
            new_obj = getattr(obj, sp_obj[0])

            if len(sp_obj) > 1:
                return self._get_attribute_trait(new_obj, sp_obj[1])
            return new_obj

        return getattr(obj, trait)



import importlib
import pkgutil


def setup_queue_handlers():
    internal_module = _get_internal_module()
    for info in pkgutil.iter_modules(internal_module.__path__):
        module_name = None
        try:
            module_name = info.name
        except AttributeError:
            module_name = info[1]
        try:
            importlib.import_module('.'.join([internal_module.__name__, module_name]))
        except ImportError:
            pass


def _get_internal_module():
    return importlib.import_module('{}.handlers'.format(__name__))



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import logging

from functools import wraps

from flask.globals import request
from flask.wrappers import Response

from rapid.ci.data.github_dal import GithubHelper
from rapid.lib import json_response, HttpException
from rapid.lib.constants import ModuleConstants
from rapid.lib.framework.injectable import Injectable
from rapid.lib.modules import WorkflowModule

logger = logging.getLogger("rapid")


class GithubController(Injectable):
    __injectables__ = {'github_helper': GithubHelper,
                       ModuleConstants.WORKFLOW_MODULE: WorkflowModule,
                       'rapid_config': 'rapid_config'}

    def __init__(self, github_helper, rapid_config, workflow_module, flask_app):
        """

        :param github_helper:
        :type github_helper: GithubHelper
        :param rapid_config:
        :type rapid_config: master.MasterConfiguration.MasterConfiguration
        :param workflow_module:
        :type workflow_module: WorkflowModule
        :param flask_app:
        :type flask_app: flask.app.Flask
        """
        self.flask_app = flask_app
        self.github_helper = github_helper
        self.rapid_config = rapid_config
        self.workflow_module = workflow_module

    def register_url_rules(self, flask_app):
        flask_app.add_url_rule("/api/github/webhooks", "github_webooks", self._github_check_wrapper(self.process_webhooks), methods=["POST"])
        flask_app.add_url_rule("/api/github/webhooks/<path:pipeline_id>", "github_webook_pipeline", self._github_check_wrapper(self.process_webhooks_pipeline), methods=['POST'])

    def _github_check_wrapper(self, func):
        @wraps(func)
        def decorated_view(*args, **kwargs):
            if 'X-Hub-Signature' in request.headers \
                    and GithubHelper.is_valid_request(request.headers['X-Hub-Signature'], self.rapid_config.github_webhooks_key, request.data):
                return func(*args, **kwargs)
            return Response("Not authorized", status=401)
        return decorated_view

    def __process_request(self):
        request_json = request.get_json()
        return request_json

    @json_response()
    def process_webhooks_pipeline(self, pipeline_id):
        request_json = self.__process_request()
        pipeline_instance = self.workflow_module.start_pipeline_by_id(pipeline_id, self._build_request_parameters(request_json))

        if pipeline_instance:
            return pipeline_instance

        raise HttpException({'message': 'Something went wrong'}, code=500)

    @json_response()
    def process_webhooks(self):
        # perform operation
        request_json = self.__process_request()

        pipeline_instance = self.workflow_module.start_pipeline_via_repo(request_json['repository']['clone_url'],
                                                                         self._build_request_parameters(request_json))

        if pipeline_instance:
            return pipeline_instance
        raise HttpException({'message': 'Something went wrong'}, code=500)

    def _build_request_parameters(self, request_json):
        parameters = {'branch': GithubHelper.get_branch_from_ref(request_json['ref']) if 'ref' in request_json else None}
        parameters.update(self._load_default_parameters(request_json))

        return {'parameters': parameters}

    def _load_default_parameters(self, request_json):
        parameters = {}
        if self.rapid_config and self.rapid_config.github_default_parameters is not None:
            for parameter_string in self.rapid_config.github_default_parameters.split('\n'):
                try:
                    tmp = parameter_string.strip().split(':')
                    parameter = tmp[0]
                    key = tmp[1]
                    value = self._get_json_value(request_json, key)
                    if parameter and value:
                        parameters[parameter] = self._get_json_value(request_json, key)
                except Exception:  # pylint: disable=broad-except
                    pass
        return parameters

    def _get_json_value(self, request_json, key):
        if '.' in key:
            tmp = key.split('.', 1)
            if tmp[0] in request_json:
                return self._get_json_value(request_json[tmp[0]], tmp[1])
            return None

        return request_json[key] if key in request_json else None



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

action_configs

Revision ID: cdb13c1069ee
Revises: 86beaee286d5
Create Date: 2019-09-25 10:55:52.728988

"""

# revision identifiers, used by Alembic.
revision = 'cdb13c1069ee'
down_revision = '86beaee286d5'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('action_configs',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('action_id', sa.Integer(), nullable=False),
    sa.Column('configuration', sa.Text(), nullable=False),
    sa.ForeignKeyConstraint(['action_id'], ['actions.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_action_configs_action_id'), 'action_configs', ['action_id'], unique=False)
    op.create_index(op.f('ix_action_configs_id'), 'action_configs', ['id'], unique=False)
    op.create_table('action_instance_configs',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('action_instance_id', sa.Integer(), nullable=False),
    sa.Column('configuration', sa.Text(), nullable=False),
    sa.ForeignKeyConstraint(['action_instance_id'], ['action_instances.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_action_instance_configs_action_instance_id'), 'action_instance_configs', ['action_instance_id'], unique=False)
    op.create_index(op.f('ix_action_instance_configs_id'), 'action_instance_configs', ['id'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_action_instance_configs_id'), table_name='action_instance_configs')
    op.drop_index(op.f('ix_action_instance_configs_action_instance_id'), table_name='action_instance_configs')
    op.drop_table('action_instance_configs')
    op.drop_index(op.f('ix_action_configs_id'), table_name='action_configs')
    op.drop_index(op.f('ix_action_configs_action_id'), table_name='action_configs')
    op.drop_table('action_configs')
    # ### end Alembic commands ###



import logging
import datetime
import random

from requests import ConnectTimeout, ReadTimeout, ConnectionError

from rapid.lib.constants import StatusConstants
from rapid.lib.framework.injectable import Injectable
from rapid.lib.store_service import StoreService
from rapid.master.communicator.master_communicator import MasterCommunicator
from rapid.workflow.action_instances_service import ActionInstanceService
from rapid.workflow.queue_handlers.queue_handler import QueueHandler
from rapid.workflow.queue_handlers.queue_handler_register import register_queue_handler

logger = logging.getLogger('rapid')


@register_queue_handler
class StandardQueueHandler(QueueHandler, Injectable):
    __injectables__ = {'rapid_config': None,
                       'action_instance_service': ActionInstanceService,
                       'flask_app': None}

    def process_work_request(self, work_request, clients):
        """
        1. Find a client that can do the work based off the label
          1a. If no label, any client that takes something that is not restricted to only its type
          1b. If label, only clients that that have at least that label, or only restricted to that type
          1c. First client to respond
        2. Save who was assigned the work
          2a. Store IP and set status to INPROGRESS
        3. Send the work to the client
          3a. If the client fails, unassign the work
        """
        clients_array = list(clients.values())
        random.shuffle(clients_array)
        pages = MasterCommunicator.find_available_clients(clients_array, work_request.grain, self.rapid_config.verify_certs)
        for client in pages:
            if client:
                if hasattr(client, 'sleep') and client.sleep:
                    continue
                try:
                    self.action_instance_service.edit_action_instance(work_request.action_instance_id, {"status_id": StatusConstants.INPROGRESS,
                                                                                                        "start_date": datetime.datetime.utcnow(),
                                                                                                        "assigned_to": "{}:{}".format(client.ip_address, client.port)})
                    try:
                        response = client.send_work(work_request, self.rapid_config.verify_certs)
                        if response.status_code == 423:
                            client.sleep = True
                            raise Exception("Client was busy, can't take work.")
                        if response.status_code == 201:
                            if 'X-Exclude-Resource'.lower() in response.headers:
                                client.sleep = True
                        else:
                            logger.info("Client didn't respond right: {} returned {}".format(client.ip_address, response.status_code))

                        break  # Break, we sent work to the other client
                    except Exception as exception:
                        logger.error("Could not send work to worker: [{}]".format(str(exception)))
                        self.action_instance_service.edit_action_instance(work_request.action_instance_id, {"status_id": StatusConstants.READY,
                                                                                                            "start_date": None,
                                                                                                            "assigned_to": None})
                except ConnectTimeout:
                    # Should reset, this is a problem, server not there.
                    self.action_instance_service.edit_action_instance(work_request.action_instance_id, {"status_id": StatusConstants.READY,
                                                                                                        "start_date": None,
                                                                                                        "assigned_to": None})
                except ReadTimeout as read_timeout:
                    logger.error(read_timeout)
                    logger.error(client.get_work_uri())
                except ConnectionError as error:
                    logger.error(error)
                    logger.error(client.get_work_uri())
                except Exception as exception:
                    logger.error(exception)
                    logger.error(client.get_work_uri())
        pages = None
        clients_array = None

    def __init__(self, rapid_config, action_instance_service, flask_app):
        """
        :param rapid_config:
        :type rapid_config: rapid.master.master_configuration.MasterConfiguration
        :param action_instance_service:
        :type action_instance_service: rapid.workflow.action_instance_service.ActionInstanceService
        :type flask_app: Flask
        """
        super(StandardQueueHandler, self).__init__(rapid_config)
        self.action_instance_service = action_instance_service
        self.flask_app = flask_app

    def can_process_work_request(self, work_request):
        return len(self._get_grain_type_split(work_request.grain)) < 2

    def can_process_action_instance(self, action_instance):
        return len(self._get_grain_type_split(action_instance['grain'])) < 2

    def process_action_instance(self, action_instance, clients):  # type: (dict, list) -> bool
        reset_action_instance = False
        if ':' not in action_instance['assigned_to']:
            logger.info("Action Instance {} assigned without port: {}".format(action_instance['id'], action_instance['assigned_to']))
            reset_action_instance = True
        else:
            ip_address, port = action_instance['assigned_to'].split(':')  # pylint: disable=unused-variable
            if ip_address in clients:
                client = clients[ip_address]
                reset_action_instance = MasterCommunicator.is_still_working_on(action_instance['id'], client,
                                                                               self.rapid_config.verify_certs) is False

        if reset_action_instance and not StoreService.is_completing(action_instance['id']):
            if self.action_instance_service.reset_action_instance(action_instance['id'], check_status=True):
                logger.info("Resetting Action Instance:{} was assigned to: {}".format(action_instance['id'], action_instance['assigned_to']))
        return True

    def cancel_worker(self, action_instance):  # type: (dict) -> bool
        for client in StoreService.get_clients(self.flask_app).values():
            if client.get_uri() == action_instance['assigned_to']:
                client.cancel_work(action_instance['id'], self.flask_app.rapid_config.verify_certs)
                break
        return True



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
try:
    import simplejson as json
except ImportError:
    import json

from sqlalchemy.sql.expression import desc

from rapid.lib.exceptions import DatabaseException, InvalidObjectException
from rapid.lib import get_db_session
from rapid.ci.data.models import Commit, CommitParameters, Version, PipelineInstanceCommit, Vcs


class CommitDal(object):
    def get_by_identifier(self, commit_identifier):
        for session in get_db_session():
            try:
                commit = session.query(Commit).filter(Commit.commit_identifier == commit_identifier).first()
                return commit.serialize
            except AttributeError:
                pass
        return None

    def create_version_for_commit(self, commit_identifier, data_version):
        for session in get_db_session():
            commit = session.query(Commit).filter(Commit.commit_identifier == commit_identifier).first()
            if commit:
                version = session.query(Version).filter(Version.name == data_version).first()
                if version:
                    raise DatabaseException("Version already exists", code=400)

                version = Version(commit_id=commit.id, name=data_version)
                session.add(version)
                session.commit()
                return version.serialize()
        raise DatabaseException("Invalid commit", code=404)

    def get_versions(self, commit_identifier, latest_only=False):
        for session in get_db_session():
            commit = session.query(Commit).filter(Commit.commit_identifier == commit_identifier).first()
            if commit:
                versions = []
                query = session.query(Version).filter(Version.commit_id == commit.id).order_by(desc(Version.date_created))
                if latest_only:
                    version = query.first()
                    if version is not None:
                        return version.serialize

                    return None

                for version in query.all():
                    versions.append(version.serialize)
                return versions

        raise DatabaseException("Invalid commit identifier")

    def _get_or_create_commit(self, commit_identifier, session, vcs_id):

        commit = session.query(Commit).filter(Commit.commit_identifier == commit_identifier).first()

        if not commit:
            commit = Commit(commit_identifier=commit_identifier, vcs_id=vcs_id)
            session.add(commit)
        return commit

    def create_git_commit(self, commit_identifier, vcs_id, additional_info=None, pipeline_instance_id=None, in_session=None):
        if in_session is None:
            for session in get_db_session():
                self._create_git_commit(commit_identifier, additional_info, pipeline_instance_id, session, True, vcs_id)
        else:
            self._create_git_commit(commit_identifier, additional_info, pipeline_instance_id, in_session, False, vcs_id)

    def _create_git_commit(self, commit_identifier, additional_info, pipeline_instance_id, session, should_commit, vcs_id):
        if commit_identifier:
            commit = self._get_or_create_commit(commit_identifier, session, vcs_id)
            if additional_info:
                for key, value in additional_info.items():
                    parameter = CommitParameters(commit_id=commit.id, name=key, value=value)
                    session.add(parameter)

            if pipeline_instance_id:
                session.flush()
                session.add(PipelineInstanceCommit(pipeline_instance_id=pipeline_instance_id, commit_id=commit.id))

            if should_commit:
                session.commit()
            return commit.serialize
        raise InvalidObjectException("Commit Identifier required.")

    def get_vcs_by_repo_name(self, repo_name):
        for session in get_db_session():
            vcs = session.query(Vcs).filter(Vcs.name == repo_name).first()
            return vcs.serialize() if vcs is not None else None

    def get_vcs_by_pipeline_id(self, pipeline_id, in_session=None):
        if in_session:
            vcs = in_session.query(Vcs).filter(Vcs.pipeline_id == pipeline_id).first()
            return vcs.serialize() if vcs is not None else None

        for session in get_db_session():
            vcs = session.query(Vcs).filter(Vcs.pipeline_id == pipeline_id).first()
            return vcs.serialize() if vcs is not None else None



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=broad-except
import logging
from abc import ABCMeta, abstractmethod
import os

import re

from rapid.lib.utils import deep_merge

logger = logging.getLogger("rapid")


class QaTestFile(object):
    __metaclass__ = ABCMeta

    AREA = "area"
    FEATURE = "feature"
    BP = "behavior_point"

    def __init__(self, filename, rootdir):
        self._tests = []
        self._filename = filename
        self._current_settings = {}
        self._found_tests = False
        self._areas = {}
        self._current_test = {}
        self._rootdir = rootdir

    @abstractmethod
    def is_reversed(self):
        raise Exception("Should implement this.")

    @property
    @abstractmethod
    def regex(self):
        raise Exception("Should implement regex.")

    def get_results(self):
        return self._areas

    def parse_file(self, name_map=None):
        with open(self._filename) as file:
            for line in file.readlines():
                try:
                    action = self.parse_action(line)
                    if action:
                        try:
                            getattr(self, "handle_{}".format(action))(line)
                        except Exception:
                            pass
                    else:
                        self.handle_test(line, name_map)
                except Exception:
                    import traceback
                    traceback.print_exc()

        if self.is_reversed() and self._current_settings:
            self._finalize_test(self._current_test)

    def parse_action(self, line):
        tmp = self._strip_comment_characters(line).lstrip()
        if tmp.startswith('rapid-') or tmp.startswith('@rapid-'):
            _match = re.findall('rapid-([a-zA-Z]+)', tmp)
            if _match and len(_match) >= 1:
                return _match[0]
        return None

    def handle_unit(self, line):
        self._current_settings['level'] = 'unit'
        self.parse_test_levels(re.split('rapid-unit:?', line)[1].strip())

    def handle_integration(self, line):
        self._current_settings['level'] = 'integration'
        self.parse_test_levels(re.split('rapid-integration:?', line)[1].strip())

    def handle_selenium(self, line):
        self._current_settings['level'] = 'selenium'
        self.parse_test_levels(re.split('rapid-selenium:?', line)[1].strip())

    def handle_tags(self, line):
        self._current_settings['tags'] = re.split('rapid-tags:?', line)[1].strip().split()

    def parse_test_levels(self, string):
        """
        :param string: 
        :type string: str 
        :return: 
        :rtype: 
        """
        split_char = ':'
        if '|' in string:
            split_char = '|'
        (area, feature, behavior_point) = string.split(split_char, 2)
        self._current_settings[self.AREA] = area
        self._current_settings[self.FEATURE] = feature
        self._current_settings[self.BP] = behavior_point

    def _strip_comment_characters(self, line):
        tmp = line.lstrip().rstrip("\n")
        return tmp.lstrip(self._get_comment_characters())

    @abstractmethod
    def _get_comment_characters(self):
        yield

    def _get_behavior_point_map(self, area, feature, behavior_point):
        area_map = self.__check_area(area)
        feature_map = self.__check_feature(area_map, feature)
        return self.__check_behavior_points(feature_map, behavior_point)

    def __check_area(self, area):
        if not self._areas.get(area, None):
            self._areas[area] = {}
        return self._areas[area]

    def __check_feature(self, area_map, feature):
        if not area_map.get(feature, None):
            area_map[feature] = {}
        return area_map.get(feature)

    def __check_behavior_points(self, feature_map, behavior_point):
        if not feature_map.get(behavior_point, None):
            feature_map[behavior_point] = []
        return feature_map[behavior_point]

    def handle_test(self, line, name_map=None):
        if self.regex:
            groups = re.findall(self.regex, line)
            if groups:
                name = groups[0]
                if name_map is not None and name in name_map:
                    name = name_map[name]

                test = {"name": name, "current_settings": dict(self._current_settings) if not self.is_reversed() else {}}
                self._tests.append(test)
                if self.is_reversed():
                    if self._current_test:
                        self._finalize_test(self._current_test)

                    self._current_test = test
                    self._current_settings = test['current_settings']
                else:
                    self._finalize_test(test)
                    self._current_settings = {}
                self._found_tests = True

    def _finalize_test(self, test):
        if test['current_settings'].get('area', None):
            _bp = self._get_behavior_point_map(test['current_settings'].get(self.AREA, None),
                                               test['current_settings'].get(self.FEATURE, None),
                                               test['current_settings'].get(self.BP, None))

            _bp.append(test)

    def printout(self):
        if self._areas:
            for area, area_map in list(self._areas.items()):
                for feature, feature_map in list(area_map.items()):
                    for _bp, tests in list(feature_map.items()):
                        print("{}:{}:{}".format(area, feature, _bp))
                        for test in tests:
                            print("      - ({}) {} -- [{}]".format(test['current_settings'].get('level'), test['name'],
                                                                   ",".join(test['current_settings'].get('tags', []))))


class PythonFile(QaTestFile):
    def _get_comment_characters(self):
        return "#"

    _CUSTOM_RE = re.compile(r'\s{1,}def (test[\S]*)\(.*$')

    @property
    def regex(self):
        return self._CUSTOM_RE

    def is_reversed(self):
        return True


class PHPFile(QaTestFile):
    def _get_comment_characters(self):
        return "/*"

    _CUSTOM_RE = re.compile(r'\s{1,}function (test[\S]*)\(.*$')

    @property
    def regex(self):
        return self._CUSTOM_RE

    def is_reversed(self):
        return False


class JSFile(QaTestFile):
    def _get_comment_characters(self):
        return "/*"

    _CUSTOM_RE = re.compile(r'\s{1,}it\([\'"`]([^\'`"]*)')

    @property
    def regex(self):
        return self._CUSTOM_RE

    def is_reversed(self):
        return False


class FileTestFactory(object):
    @staticmethod
    def create_file_test(filename, rootdir):
        """
        :return FileTest
        """
        file_type = filename.split('.')[-1]
        if file_type == 'php':
            return PHPFile(filename, rootdir)
        if file_type == 'py':
            return PythonFile(filename, rootdir)
        if file_type == 'js':
            return JSFile(filename, rootdir)
        return None


def process_directory(workspace, dirname, cmdline=False, name_map=None):
    results = {}
    for root, dirs, files in os.walk(dirname):  # pylint: disable=unused-variable
        for filename in files:
            try:
                testfile = FileTestFactory.create_file_test("{}/{}".format(root, filename), workspace)
                if testfile:
                    testfile.parse_file(name_map)
                    deep_merge(results, testfile.get_results())
            except Exception as exception:
                logger.error("Was Unable to process the directory defined. {}".format(exception))
    if cmdline:
        __print_results(results)
    return results


def print_results(results):
    __print_results(results)


def __print_results(areas):
    for area, area_map in areas.items():
        print(area)
        for feature, feature_map in area_map.items():
            print(("  {}".format(feature)))
            for _bp, tests in feature_map.items():
                print(("   {}".format(_bp)))
                # print "{}:{}:{}".format(area, feature, bp)
                for test in tests:
                    print("      - ({}) {} -- [{}]".format(test['current_settings'].get('level'), test['name'],
                                                           ",".join(test['current_settings'].get('tags', []))))
        print("")


def register_ioc_globals(flask_app):  # pylint: disable=unused-argument
    pass


def configure_module(flask_app):  # pylint: disable=unused-argument
    pass



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import logging
import calendar
import re
from datetime import datetime

from sqlalchemy import Column, Integer
from sqlalchemy.ext.declarative import declared_attr

logger = logging.getLogger("rapid")
# pylint: disable=no-member


class BaseModel(object):

    @declared_attr
    def __tablename__(self):
        name = "_".join(re.findall('[A-Z]{1}[a-z]*', self.__name__))
        name = name.lower()
        if name.endswith('s') and re.search("[aeiou]s$", name):
            name = "{}es".format(name)
        elif name.endswith('y'):
            name = name[:-1] + "ies"
        elif not name.endswith('s'):
            name = "%s%s" % (name, "s"[2 == 1:])
        return name

    __table_args__ = {'mysql_engine': 'InnoDB'}

    id = Column(Integer, primary_key=True, index=True)

    def __repr__(self):
        return "<Model {}>".format(self.__class__)

    def __relationships__(self):
        """
        Return a list of relationships name which are not as a backref
        name in model
        """
        from sqlalchemy.inspection import inspect
        relationships = []
        for item in inspect(self.__class__).relationships:
            relationships.append("{}".format(item).split('.')[1])
        return relationships

    def serialize(self, allowed_children=None, previous_relationship=None):
        serialized = Odict()
        relationships = self.__relationships__()
        fields = [x for x in dir(self) if not x.startswith('_') and x not in ['metadata', 'get_parameters_dict']]
        self._serialize_fields(serialized, fields, previous_relationship, allowed_children, relationships)
        return serialized

    def _serialize_fields(self, serialized, fields, previous_relationship, allowed_children, relationships):
        for field in fields:
            key_filter = self.__tablename__ if previous_relationship is None else previous_relationship
            allowed_filter = dict(allowed_children) if allowed_children else {}
            if field == 'serialize' \
                    or field == 'query' \
                    or field == 'query_class' \
                    or field == 'convert_to_instance' \
                    or field in relationships:
                if allowed_children and field in relationships:
                    allowed_filter = self.__allowed_iteration(allowed_children, key_filter, field)
                    if field not in allowed_filter:
                        if 'fields' in allowed_filter and field in allowed_filter['fields']:
                            pass
                        else:
                            continue
                else:
                    continue
            value = self.__getattribute__(field)

            if isinstance(value, list):
                value = [item.serialize(allowed_filter, previous_relationship=field) for item in value]
            elif isinstance(value, datetime):
                value = int(calendar.timegm(value.utctimetuple()))
            elif hasattr(value, 'serialize'):
                value = value.serialize(allowed_filter, previous_relationship=field)

            serialized[field] = value
            setattr(serialized, field, value)

    def __allowed_iteration(self, allowed_children, key_filter, field):
        allowed_combine = {}
        if key_filter in allowed_children and field in allowed_children[key_filter]:
            allowed_combine.update({key_filter: allowed_children[key_filter]})
            allowed_combine.update({"fields": allowed_children[key_filter]})

            if field in allowed_children:
                allowed_combine.update({field: allowed_children[field]})
                for field2 in allowed_children[field]:
                    allowed_combine.update(self.__allowed_iteration(allowed_children, field, field2))
        return allowed_combine


class Odict(dict):
    pass



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
try:
    import simplejson as json
except ImportError:
    import json

import os
from os.path import basename

import time
import pickle
import logging
import threading

from flask.globals import request
from flask.wrappers import Response

from rapid.client.communicator.client_communicator import ClientCommunicator
from rapid.lib import api_key_required
from rapid.lib.store_service import StoreService
from rapid.lib.utils import UpgradeUtil
from rapid.lib.work_request import WorkRequest
from rapid.client.executor import Executor
from ...lib.base_controller import BaseController
from ...lib.version import Version

logger = logging.getLogger("rapid")
# pylint: disable=broad-except


class WorkController(BaseController):
    executors = []

    def __init__(self):
        self.app = None

    def register_url_rules(self, flask_app):
        """
        This is the entry point for the controller. All URL ruls are add_url_rule(d) here.
        :param flask_app: The app you want to add a url to
        :return: void
        """
        flask_app.add_url_rule('/work/request', 'work_request', api_key_required(self.work_request))
        flask_app.add_url_rule('/work/execute', 'work_execute', api_key_required(self.work_execute), methods=['POST'])
        flask_app.add_url_rule('/work/cancel/<path:action_instance_id>', 'work_cancel', api_key_required(self.work_cancel), methods=['POST'])
        self.app = flask_app

    def work_request(self):
        current_work = StoreService.get_executors()
        current_work = current_work + self.__get_quarantined_items()
        work = {'current_work': current_work,
                'hostname': self.app.rapid_config.hostname}
        if self.can_work_on() and self.check_version(request):
            return Response(json.dumps(work), content_type='application/json', headers={Version.HEADER: Version.get_version()})
        return Response(json.dumps(work), status=423, content_type='application/json', headers={Version.HEADER: Version.get_version()})

    def __get_quarantined_items(self):
        items = []
        quarantine_directory = self.app.rapid_config.quarantine_directory
        if quarantine_directory:
            communicator = None
            if not os.path.isdir(quarantine_directory):
                os.makedirs(quarantine_directory)

            for item in os.listdir(quarantine_directory):
                if item in ['.', '..']:
                    continue
                try:
                    items.append({'action_instance_id': int(item), 'pid': 'quarantined'})
                    if communicator is None:
                        communicator = ClientCommunicator(self.app.rapid_config.master_uri,
                                                          self.app.rapid_config.quarantine_directory,
                                                          verify_certs=self.app.rapid_config.verify_certs,
                                                          get_files_auth=self.app.rapid_config.get_files_basic_auth)
                    try:
                        delete_file = False
                        with open("{}/{}".format(quarantine_directory, item), 'r') as tmp_file:
                            data = pickle.loads(tmp_file.read())
                            try:
                                status = data['status']
                                parameters = data['parameters'] if 'parameters' in data else None
                                stats = data['stats'] if 'stats' in data else None
                                results = data['results'] if 'results' in data else None

                                communicator.send_done(int(item), status, parameters, stats, results, logger, headers={'X-Rapid-Quarantine': 'true'})
                                delete_file = True
                            except Exception:
                                import traceback
                                traceback.print_exc()

                        if delete_file:
                            try:
                                os.remove("{}/{}".format(quarantine_directory, item))
                            except Exception:
                                logger.error("Couldn't remove.")
                    except Exception:
                        import traceback
                        traceback.print_exc()

                except Exception:
                    pass
        return items

    def can_work_on(self, work=None):
        """
        This checks the work queue and what is being run at the moment.
        :param work
        :type work dict
        :return: True|False
        """
        executors = StoreService.get_executors()
        return len(executors) < self.app.rapid_config.executor_count and not self.currently_running(work)

    def currently_running(self, work):
        pid_exists = False

        try:
            if work['action_instance_id'] is not None:
                pid_exists = StoreService.check_for_pidfile(work['action_instance_id'])
                if pid_exists is not None:
                    logger.info("Request was sent, but was already running, ignoring for [{}]".format(work['action_instance_id']))
        except Exception:
            pass

        return pid_exists

    def check_version(self, check_request):
        if Version.HEADER in check_request.headers:
            if check_request.headers[Version.HEADER] == self.get_version():
                return True
            if not StoreService.is_updating(self.app):
                StoreService.set_updating(self.app)
                thread = threading.Thread(target=self._perform_upgrade, args=(check_request.headers[Version.HEADER],))
                thread.daemon = True
                thread.start()

                return False
        return False

    def _perform_upgrade(self, new_version):
        logger.info("Performing upgrade: {}".format(new_version))

        executors = self._sleep_for_executors()
        self._start_upgrade(new_version, executors)

    def _start_upgrade(self, new_version, executors):
        if not executors:
            UpgradeUtil.upgrade_version(new_version, self.app.rapid_config)
        else:
            logger.info("Waiting for executors...")
            StoreService.set_updating(self.app, False)

    @staticmethod
    def _sleep_for_executors(seconds_to_sleep=10, count_limit=10):
        executors = StoreService.get_executors()
        count = 0
        while executors:
            time.sleep(seconds_to_sleep)
            count += 1
            if count >= count_limit:
                break
            executors = StoreService.get_executors()
        return executors

    def get_version(self):
        return Version.get_version()

    def work_execute(self):
        try:
            if self.can_work_on(request.get_json()):
                self.start_work()
                executors = StoreService.get_executors()
                headers = {}
                if len(executors) + 1 >= self.app.rapid_config.executor_count:
                    headers["X-Exclude-Resource"] = 'true'
                return Response(json.dumps({"message": "Work started"}), status=201, content_type="application/json", headers=headers)
        except Exception as exception:
            logger.error(exception)
            return Response(json.dumps({'message': str(exception)}), status=422, content_type='application/json')

        return Response(json.dumps({"message": "Cannot execute work at this time."}), status=423, content_type='application/json')

    def start_work(self):
        executor = Executor(WorkRequest(request.get_json()),
                            self.app.rapid_config.master_uri,
                            workspace=self.app.rapid_config.workspace,
                            quarantine=self.app.rapid_config.quarantine_directory,
                            verify_certs=self.app.rapid_config.verify_certs,
                            rapid_config=self.app.rapid_config)
        executor.verify_work_request()
        executor.start()

    def work_cancel(self, action_instance_id):
        pid_file = StoreService.check_for_pidfile(action_instance_id)
        if pid_file is not None:
            try:
                base_name = basename(pid_file)
                os.kill(int(base_name.split('-')[-1]), 9)
                return Response(json.dumps({"message": "Killed process."}), 200)
            except Exception:
                pass

        return Response(json.dumps({"message": "Failed to kill process"}), 501)




"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.lib.framework.ioc import IOC


def register_ioc_globals(flask_app):  # pylint: disable=unused-argument
    from rapid.workflow.action_service import ActionService
    from rapid.workflow.workflow_service import WorkflowService
    from rapid.lib.queue_handler_constants import QueueHandlerConstants
    from rapid.workflow.queue_handlers import setup_queue_handlers

    setup_queue_handlers()

    IOC.register_global('queue_constants', IOC.get_class_instance(QueueHandlerConstants))
    IOC.register_global('workflow_module', IOC.get_class_instance(WorkflowService))
    IOC.register_global('action_module', IOC.get_class_instance(ActionService))


def configure_module(flask_app):
    from rapid.workflow.api_controller import APIRouter
    router = IOC.get_class_instance(APIRouter)
    router.register_url_rules(flask_app)





"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import UniqueConstraint
from sqlalchemy import Column, String, ForeignKey, Integer, DateTime, Boolean, Text

from rapid.lib import get_declarative_base
from rapid.master.data.database.models.base.active_model import ActiveModel
from rapid.master.data.database.models.base.base_model import BaseModel
# pylint: disable=no-member,too-few-public-methods
Base = get_declarative_base()


class QaTest(BaseModel, Base):
    name = Column(String(1000), nullable=False)
    active = Column(Boolean, default=True)
    qa_test_type_id = Column(Integer, ForeignKey('qa_test_types.id'), nullable=True, index=True)


class QaStatusSummary(BaseModel, Base):
    action_instance_id = Column(Integer, ForeignKey('action_instances.id'), nullable=False, index=True)
    status_id = Column(Integer, ForeignKey('statuses.id'), nullable=False, index=True)
    count = Column(Integer)

    status = relationship('Status')
    action_instance = relationship("ActionInstance", backref="status_summary")


class QaTestHistory(BaseModel, Base):
    test_id = Column(Integer, ForeignKey('qa_tests.id'), nullable=False, index=True)
    pipeline_instance_id = Column(Integer, ForeignKey('pipeline_instances.id'), nullable=False, index=True)
    action_instance_id = Column(Integer, ForeignKey('action_instances.id'), nullable=False, index=True)
    duration = Column(Integer)
    status_id = Column(Integer, ForeignKey('statuses.id'), index=True)
    date_created = Column(DateTime(), nullable=False, default=datetime.datetime.utcnow)

    test = relationship('QaTest')
    pipeline_instance = relationship('PipelineInstance')
    action_instance = relationship('ActionInstance', backref="qa_test_histories")
    status = relationship('Status')
    stacktrace = relationship("Stacktrace", single_parent=True, backref="qa_test_history", cascade="all,delete,delete-orphan")


class Stacktrace(BaseModel, Base):
    qa_test_history_id = Column(Integer, ForeignKey('qa_test_histories.id'), nullable=False, index=True, unique=True)
    stacktrace = Column(Text)


class QaProduct(BaseModel, Base):
    name = Column(String(250), nullable=False, index=True)
    vcs_id = Column(Integer, ForeignKey('vcs.id'), nullable=False, index=True)

    qa_areas = relationship('QaArea', backref="qa_product", order_by="asc(QaArea.name)")
    vcs = relationship('Vcs', backref='qa_product')


class QaArea(BaseModel, Base):
    name = Column(String(250), nullable=False, index=True)
    product_id = Column(Integer, ForeignKey("qa_products.id"), nullable=False, index=True)


class QaFeature(BaseModel, Base):
    name = Column(String(250), nullable=False, index=True)
    product_id = Column(Integer, ForeignKey("qa_products.id"), nullable=False, index=True)


class QaBehaviorPoint(BaseModel, Base):
    name = Column(String(400), nullable=False, index=True)
    product_id = Column(Integer, ForeignKey("qa_products.id"), nullable=False, index=True)


class QaTestMapping(BaseModel, Base):
    area_id = Column(Integer, ForeignKey("qa_areas.id"), nullable=False, index=True)
    feature_id = Column(Integer, ForeignKey('qa_features.id'), index=True)
    behavior_id = Column(Integer, ForeignKey('qa_behavior_points.id'), nullable=False, index=True)
    test_id = Column(Integer, ForeignKey('qa_tests.id'), nullable=False)

    area = relationship('QaArea')
    feature = relationship('QaFeature')
    behavior_point = relationship('QaBehaviorPoint')
    test = relationship('QaTest')

    __table_args__ = (UniqueConstraint('area_id', 'feature_id', 'behavior_id', 'test_id', name="uix_qatm"),)


class QaTestMapTag(BaseModel, Base):
    qa_testmap_id = Column(Integer, ForeignKey('qa_test_mappings.id'), nullable=False, index=True)
    tag_id = Column(Integer, ForeignKey('tags.id'), nullable=False, index=True)


class Tag(BaseModel, Base):
    name = Column(String(100), nullable=False, index=True)


class QaTestType(ActiveModel, Base):
    pass



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import logging
import os
import threading
import time

try:
    import simplejson as json
except ImportError:
    import json

from flask import Flask

from rapid.lib import is_primary_worker, setup_logging, setup_status_route
from .parsers import load_parsers
from ..lib import setup_config_from_file
from .communicator.client_communicator import ClientCommunicator
from .controllers import register_controllers

app = Flask("rapidci_client")
app.rapid_config = {'_is': 'client'}
logger = logging.getLogger("rapid")


@app.errorhandler(500)
def internal_error(exception):
    response = json.dumps(exception.to_dict())
    response.status_code = exception.status_code
    response.content_type = 'application/json'
    return response


def setup_logger(flask_app):
    setup_logging(flask_app)


def load_extensions(flask_app):
    from rapid.lib.framework.ioc import IOC
    from rapid.extensions.extension_loader import ExtensionLoader
    extension_loader = IOC.get_class_instance(ExtensionLoader)
    extension_loader.load_extensions(flask_app)


def configure_application(flask_app, args):
    setup_status_route(flask_app)
    setup_config_from_file(flask_app, args)
    setup_logger(flask_app)
    load_parsers()
    register_controllers(flask_app)
    if is_primary_worker() and not args.run and not args.upgrade:
        setup_client_register_thread()
        clean_workspace()

    if args.mode_logging:
        from rapid.lib.log_server import LogServer
        log_server = LogServer(args.log_dir)
        log_server.configure_application(flask_app)

    load_extensions(flask_app)

def clean_workspace():
    try:
        import shutil
        if os.path.isdir(app.rapid_config.workspace):  # pylint: disable=no-member
            shutil.rmtree(app.rapid_config.workspace)  # pylint: disable=no-member
        os.mkdir(app.rapid_config.workspace)  # pylint: disable=no-member
    except Exception:  # pylint: disable=broad-except
        pass


def _registration_thread():
    communicator = ClientCommunicator(app.rapid_config.master_uri, app.rapid_config.quarantine_directory, app, app.rapid_config.verify_certs)  # pylint: disable=no-member
    while True:
        communicator.register(app.rapid_config)
        time.sleep(app.rapid_config.registration_rate)  # pylint: disable=no-member


def setup_client_register_thread():
    logger.info("Setting up client register thread")
    thread = threading.Thread(target=_registration_thread)
    thread.daemon = True
    thread.start()


def run_action_instance(action_instance_id):
    # type: (int) -> None
    from rapid.client.action_instance_runner import ActionInstanceRunner
    runner = ActionInstanceRunner(app.rapid_config)
    runner.run_action_instance(app, action_instance_id)


def upgrade_rapid():
    from rapid.client.client_upgrader import ClientUpgrader
    upgrader = ClientUpgrader(app.rapid_config)
    upgrader.upgrade(app)



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

events_model

Revision ID: bf363c3a9ef0
Revises: e66b5abaa8c2
Create Date: 2017-10-20 17:25:24.391418

"""

# revision identifiers, used by Alembic.
from rapid.workflow.data.models import EventType

revision = 'bf363c3a9ef0'
down_revision = 'e66b5abaa8c2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('event_types',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_event_types_id'), 'event_types', ['id'], unique=False)
    op.create_table('pipeline_events',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('pipeline_id', sa.Integer(), nullable=False),
    sa.Column('event_type_id', sa.Integer(), nullable=False),
    sa.Column('conditional', sa.TEXT(), nullable=False),
    sa.Column('config', sa.TEXT(), nullable=False),
    sa.ForeignKeyConstraint(['pipeline_id'], ['pipelines.id'], ),
    sa.ForeignKeyConstraint(['event_type_id'], ['event_types.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_pipeline_events_event_type_id'), 'pipeline_events', ['event_type_id'], unique=False)
    op.create_index(op.f('ix_pipeline_events_id'), 'pipeline_events', ['id'], unique=False)
    op.create_index(op.f('ix_pipeline_events_pipeline_id'), 'pipeline_events', ['pipeline_id'], unique=False)
    ### end Alembic commands ###

    op.bulk_insert(EventType.__table__, [
        {"id": 1, "name": "Remote Webhook", "active": True}
    ])


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_pipeline_events_pipeline_id'), table_name='pipeline_events')
    op.drop_index(op.f('ix_pipeline_events_id'), table_name='pipeline_events')
    op.drop_index(op.f('ix_pipeline_events_event_type_id'), table_name='pipeline_events')
    op.drop_table('pipeline_events')
    op.drop_index(op.f('ix_event_types_id'), table_name='event_types')
    op.drop_table('event_types')
    ### end Alembic commands ###



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.ci.data.commit_dal import CommitDal
from rapid.lib.framework.injectable import Injectable
from rapid.lib.modules import CiModule


class VcsService(Injectable, CiModule):
    __injectables__ = {'commit_dal': CommitDal}

    def __init__(self, commit_dal):
        """
        :type commit_dal: :class:`rapid.ci.data.commit_dal.CommitDal`
        :return:
        """
        self.commit_dal = commit_dal

    def get_vcs_by_repo_name(self, repo_name):
        return self.commit_dal.get_vcs_by_repo_name(repo_name)

    def create_git_commit(self, commit_identifier, vcs_id, pipeline_instance_id=None, additional_parameters=None, session=None):
        """
        Create a git commit entry.
        :param commit_identifier:
        :param additional_parameters:
        :type additional_parameters: dict
        :return:
        """
        return self.commit_dal.create_git_commit(commit_identifier, vcs_id, additional_parameters, pipeline_instance_id, session)

    def create_version_for_commit(self, commit_identifier, version):
        return self.commit_dal.create_version_for_commit(commit_identifier, version)

    def get_by_identifier(self, commit_identifier):
        return self.commit_dal.get_by_identifier(commit_identifier)

    def get_vcs_by_pipeline_id(self, pipeline_id, session=None):
        return self.commit_dal.get_vcs_by_pipeline_id(pipeline_id, session)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import datetime

from sqlalchemy import Column, String, Integer, ForeignKey, desc, DateTime, Boolean
from sqlalchemy.orm import relationship

from rapid.lib import get_declarative_base
from rapid.master.data.database.models.base.base_model import BaseModel

Base = get_declarative_base()
# pylint: disable=no-member, too-few-public-methods


class Commit(BaseModel, Base):
    commit_identifier = Column(String(255), nullable=False, index=True)
    vcs_id = Column(Integer, ForeignKey('vcs.id'), nullable=False, index=True)

    pipeline_instances = relationship("PipelineInstance", backref="commit", secondary="pipeline_instance_commits", order_by=desc("created_date"))
    versions = relationship('Version', backref="commit", order_by=desc('date_created'))
    vcs = relationship('Vcs')


class CommitIntegration(BaseModel, Base):
    commit_id = Column(Integer, ForeignKey('commits.id'), nullable=False, index=True)
    integration_id = Column(Integer, ForeignKey('integrations.id'), nullable=False, index=True)


class CommitParameters(BaseModel, Base):
    commit_id = Column(Integer, ForeignKey("commits.id"), index=True)
    name = Column(String(255), nullable=False, index=True)
    value = Column(String(500), nullable=False)


class Version(BaseModel, Base):
    name = Column(String(255), nullable=False, index=True)
    commit_id = Column(Integer, ForeignKey("commits.id"))
    date_created = Column(DateTime(), nullable=False, default=datetime.datetime.utcnow)


class PipelineInstanceCommit(BaseModel, Base):
    pipeline_instance_id = Column(Integer, ForeignKey("pipeline_instances.id"), index=True, nullable=False)
    commit_id = Column(Integer, ForeignKey('commits.id'), index=True, nullable=False)


class Vcs(BaseModel, Base):
    name = Column(String(1500), nullable=False)
    repo = Column(String(500), nullable=False)
    active = Column(Boolean, default=True, nullable=False)
    vcs_type_id = Column(Integer, ForeignKey('vcs_types.id'), nullable=False)
    pipeline_id = Column(Integer, ForeignKey('pipelines.id'), index=True, nullable=True)

    vcs_type = relationship("VcsType")
    pipeline = relationship("Pipeline")


class VcsType(BaseModel, Base):
    name = Column(String(255), nullable=False)
    active = Column(Boolean, default=True, nullable=False)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=no-member
import os

from flask_sqlalchemy import SQLAlchemy

from rapid.lib import set_db


def configure_db():
    _db = SQLAlchemy()
    set_db(_db)


def configure_data_layer(flask_app):
    from rapid.lib import db as _db
    flask_app.config['SQLALCHEMY_DATABASE_URI'] = flask_app.rapid_config.db_connect_string
    flask_app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    flask_app.db = _db
    _db.init_app(flask_app)

    from rapid.master.data.database.dal import setup_dals
    setup_dals(flask_app)


def run_db_upgrades(flask_app):
    with flask_app.app_context():
        from alembic import command

        config = _get_alembic_config(flask_app)
        command.upgrade(config, 'head')


def run_db_downgrade(flask_app, revision):
    with flask_app.app_context():
        from alembic import command

        config = _get_alembic_config(flask_app)
        command.downgrade(config, revision)


def create_revision(flask_app, message=None):
    with flask_app.app_context():
        from alembic import command

        config = _get_alembic_config(flask_app)
        command.revision(config, message, autogenerate=True)


def _get_alembic_config(flask_app):
    from alembic.config import Config
    return Config(config_args={'script_location': "{}/migrations".format(os.path.dirname(__file__)),
                               'sqlalchemy.url': flask_app.rapid_config.db_connect_string})


__all__ = ['configure_data_layer', 'create_revision']





"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=broad-except

import logging
from concurrent.futures import ThreadPoolExecutor
import requests
from requests.exceptions import ConnectTimeout, ConnectionError  # pylint: disable=redefined-builtin

from rapid.lib.version import Version
from rapid.lib.communicator import Communicator


logger = logging.getLogger("rapid")


class MasterCommunicator(Communicator):

    @staticmethod
    def find_available_clients(clients, grain, verify_certs=True):
        pages = None
        with ThreadPoolExecutor(max_workers=5) as executor:
            pages = executor.map(MasterCommunicator.check_availability,
                                 [(client, verify_certs) for client in MasterCommunicator.filter_clients(clients, grain)])

        return pages

    @staticmethod
    def get_clients_working_on(clients, verify_certs=True):
        pages = None
        with ThreadPoolExecutor(max_workers=5) as executor:
            pages = executor.map(MasterCommunicator.get_availability,
                                 [(client, verify_certs) for client in clients])
        return pages

    @staticmethod
    def filter_clients(clients, grain):
        tmp = []
        for client in clients:
            if client.can_handle(grain) and not hasattr(client, 'no-longer-active'):
                tmp.append(client)
        return tmp

    @staticmethod
    def check_availability(to_check):
        client, verify_certs = to_check
        try:
            response = requests.get(client.get_availability_uri(), headers=client.get_headers(), verify=verify_certs, timeout=client.time_elapse)
            if response.status_code == 200:
                return client

            if response.status_code == 423:
                client.sleep = True
                return client
        except ConnectionError:
            setattr(client, 'no-longer-active', True)
        except Exception:
            import traceback
            traceback.print_exc()
        return None

    @staticmethod
    def get_availability(to_check):
        client, verify_certs = to_check
        try:
            response = requests.get(client.get_availability_uri(), headers=client.get_headers(), verify=verify_certs, timeout=client.time_elapse)
            version = response.headers[Version.HEADER] if Version.HEADER in response.headers else 'Unknown'
            results = {"version": version, "ip_address": client.ip_address, "grains": client.grains}
            results.update(response.json())
            return results
        except Exception:
            import traceback
            traceback.print_exc()
        return None

    @staticmethod
    def is_still_working_on(action_instance_id, client, verify_certs=True):
        is_still_working = False
        try:
            response = requests.get(client.get_availability_uri(), headers=client.get_headers(), verify=verify_certs)
            try:
                if 'current_work' in response.json():
                    for instance in response.json()['current_work']:
                        if int(action_instance_id) == int(instance['action_instance_id']):
                            is_still_working = True
                            break
                else:
                    logger.info("CURRENT WORK NOT FOUND: Client:[{}] Status[{}]\nContent:{}".format(client.ip_address, response.status_code, response.content))
            except Exception:
                is_still_working = True

        except (ConnectTimeout, requests.exceptions.Timeout):
            # Not sure if there is a network issue, don't do anything
            is_still_working = True
        except ConnectionError as exception:
            # Server is up, but port is down.
            logger.error(exception)
            is_still_working = False
        except Exception as exception:
            logger.info("is_still_working failed with: {}".format(str(exception)))
            logger.exception(exception)

        return is_still_working




import datetime
import logging
import json
import re

import boto3
from botocore.exceptions import ClientError

from rapid.lib.constants import StatusConstants
from rapid.lib.exceptions import ECSLimitReached, QueueHandlerShouldSleep, ECSConnectionError
from rapid.lib.framework.injectable import Injectable
from rapid.lib.work_request import WorkRequest
from rapid.master.master_configuration import MasterConfiguration
from rapid.workflow.action_instances_service import ActionInstanceService
from rapid.workflow.queue_handlers.container_handlers.container_handler import ContainerHandler
from rapid.workflow.queue_handlers.container_handlers.ecs_configuration import ECSConfiguration
from rapid.workflow.queue_handlers.queue_handler_register import register_queue_handler

logger = logging.getLogger('rapid')


@register_queue_handler
class ECSQueueHandler(ContainerHandler, Injectable):
    __injectables__ = {'rapid_config': None, 'action_instance_service': ActionInstanceService}
    @property
    def container_identifier(self):
        return 'ecs'

    def __init__(self, rapid_config, action_instance_service):
        """
        Parameters
        ----------
        rapid_config: MasterConfiguration
        action_instance_service: ActionInstanceService
        """
        super(ECSQueueHandler, self).__init__(rapid_config)
        self.action_instance_service = action_instance_service
        self._ecs_client = None
        self._ecs_configuration = None  # type: ECSConfiguration or None

    def process_work_request(self, work_request, clients):
        # type: (WorkRequest, list) -> bool
        # 1. Get the Task Definition
        self._get_ecs_client()  # load the client and get ready to use the taskDefinition

        task_definition = self._get_task_definition(work_request)

        if 'taskDefinition' not in task_definition:
            self._set_task_status(work_request.action_instance_id, StatusConstants.FAILED,
                                  start_date=datetime.datetime.utcnow(),
                                  end_date=datetime.datetime.utcnow())
        else:
            try:
                # 2. Run the Task
                (status_id, assigned_to) = self._run_task(task_definition)

                # 3. Verify the status in case it failed.
                end_date = None
                if status_id > StatusConstants.INPROGRESS:
                    end_date = datetime.datetime.utcnow()

                # 3. Report the status
                if end_date:
                    self._set_task_status(work_request.action_instance_id, status_id, assigned_to, end_date=end_date)
                elif status_id != StatusConstants.READY:
                    self._set_task_status(work_request.action_instance_id, status_id, assigned_to, start_date=datetime.datetime.utcnow())
            except ECSLimitReached:
                raise QueueHandlerShouldSleep('ECS Limit was reached.')
            except ECSConnectionError:
                raise QueueHandlerShouldSleep('ECS Connection issue detected.')
        return True

    def process_action_instance(self, action_instance, clients):
        # 1. Look at the assigned_to for the specific ARN
        arn = action_instance['assigned_to'].split('--ecs--')[1]
        task = self._get_ecs_client().list_tasks(cluster=self._ecs_configuration.default_task_definition['cluster'],
                                                 desiredStatus='RUNNING')

        # 2. Verify that the assigned_to ARN is running.
        if task and 'taskArns' in task:
            if arn in task['taskArns']:
                return True
            self.action_instance_service.reset_action_instance(action_instance['id'], check_status=True)

    def cancel_worker(self, action_instance):  # type: (dict) -> bool
        try:
            arn = action_instance['assigned_to'].split('--ecs--')[1]
            task = self._get_ecs_client().stop_task(cluster=self._ecs_configuration.default_task_definition['cluster'],
                                                    task=arn,
                                                    reason='Rapid canceled.')
            if 'task' in task and task['task']:
                return True

            logger.info("Failed to cancel ECS arn: {}".format(arn))
        except Exception as exception:
            logger.exception(exception)
        return False


    ########################
    # Private Methods
    ########################

    def _get_task_definition(self, work_request):
        task_definition = self._get_overridden_task_definition(work_request.configuration)
        self._inject_work_request_parameters(task_definition, work_request)
        return task_definition

    def _get_overridden_task_definition(self, configuration):
        # type: (str) -> dict
        task_definition = self._ecs_configuration.default_task_definition
        try:
            if configuration:
                for key, value in json.loads(configuration).items():
                    task_definition[key] = value
        except (ValueError, TypeError) as exception:
            logger.exception(exception)

        return task_definition

    def _inject_work_request_parameters(self, task_definition, work_request):
        # type:(dict, WorkRequest) -> None
        environment_default = [{'name': 'action_instance_id', 'value': str(work_request.action_instance_id)},
                               {'name': 'workflow_instance_id', 'value': str(work_request.workflow_instance_id)},
                               {'name': 'pipeline_instance_id', 'value': str(work_request.pipeline_instance_id)}]

        (grain_type, image) = self._get_grain_type_split(work_request.grain)
        container_overrides = self._get_task_definition_key(task_definition, 'overrides:dict.containerOverrides:list')
        container_overrides.append({"name": image.split(':')[0],
                                    "environment": environment_default})

        task_definition['taskDefinition'] = self._get_substituted_value(work_request, image.strip())
        task_definition['count'] = 1

    @staticmethod
    def _get_task_definition_key(task_definition, in_key):
        current_pointer = None
        for tmp_key in in_key.split('.'):
            try:
                (key, default_type) = tmp_key.split(':')

                try:
                    if current_pointer is not None:
                        current_pointer = current_pointer[key]
                    else:
                        current_pointer = task_definition[key]
                except KeyError:
                    if current_pointer is not None:
                        current_pointer[key] = eval('{}()'.format(default_type))
                        current_pointer = current_pointer[key]
                    else:
                        current_pointer = eval('{}()'.format(default_type))
                        task_definition[key] = current_pointer
            except ValueError:
                pass
        return current_pointer

    def _set_task_status(self, action_instance_id, status_id, assigned_to='', start_date=None, end_date=None):
        # type: (int, int, str, datetime.datetime or None, datetime.datetime or None) -> None
        assigned_to = '--ecs--{}'.format(assigned_to) if '--ecs--' not in assigned_to else assigned_to
        changes = {'status_id': status_id, 'assigned_to': assigned_to}
        if start_date:
            changes['start_date'] = start_date
        if end_date:
            changes['end_date'] = end_date
        self.action_instance_service.edit_action_instance(action_instance_id, changes)
        
    def _run_task(self, task_definition):
        ecs_client = self._get_ecs_client()
        status_id = StatusConstants.INPROGRESS
        assigned_to = ''
        try:
            response_dict = ecs_client.run_task(**task_definition)
            if response_dict['failures']:
                logger.error("Failures were found: [{}]".format(response_dict['failures']))
                if 'limit' in response_dict['failures'][0]['reason']:
                    raise ECSLimitReached(response_dict['failures'][0]['reason'])
                elif 'connect timed out' in response_dict['failures'][0]['reason']:
                    raise ECSConnectionError(response_dict['failures'][0]['reason'])
                else:
                    status_id = StatusConstants.FAILED
            elif not response_dict['tasks']:
                logger.error("No tasks were returned")
                status_id = StatusConstants.READY
            elif response_dict['tasks']:
                for task in response_dict['tasks']:
                    assigned_to = '--ecs--{}'.format(task['taskArn'])
        except (KeyError, ClientError, IndexError) as exception:
            logger.error(exception)
            status_id = StatusConstants.FAILED
        
        return status_id, assigned_to

    @staticmethod
    def _get_substituted_value(work_request, sub_string):
        if '{' in sub_string:
            try:
                for match in re.findall(r'{[\w_]*}', sub_string):
                    token = match.split('{')[1].split('}')[0]
                    try:
                        sub_string = sub_string.replace(match, '{}'.format(getattr(work_request, token)))
                    except AttributeError:
                        pass
            except (TypeError, AttributeError):
                pass
        return sub_string

    def _get_ecs_client(self):
        if self._ecs_client is None:
            self._ecs_client = self._load_aws_config()
        return self._ecs_client

    def _load_aws_config(self):
        self._ecs_configuration = ECSConfiguration(self.rapid_config.ecs_config_file)
        return boto3.client('ecs', **self._ecs_configuration.aws_credentials)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import time
import jsonpickle
from flask import request, Response

from rapid.lib import api_key_required, json_response
from rapid.lib.constants import HeaderConstants
from rapid.lib.exceptions import HttpException, VcsNotFoundException
from rapid.lib.store_service import StoreService
from rapid.lib.version import Version
from rapid.master.communicator.client import Client
from rapid.master.communicator.master_communicator import MasterCommunicator
from rapid.workflow.action_dal import ActionDal

try:
    import simplejson as json
except ImportError:
    import json

# pylint: disable=broad-except


class UtilityRouter(object):

    def __init__(self, flask_app=None):
        self.flask_app = flask_app

    def register_url_rules(self):
        self.flask_app.add_url_rule('/clients/show', 'get_clients', api_key_required(self.show_clients))
        self.flask_app.add_url_rule('/client/register', 'register_client', self.register_client, methods=["POST"])
        self.flask_app.add_url_rule('/clients/working', 'get_clients_working', api_key_required(self.working_clients), methods=['GET'])
        self.flask_app.add_url_rule('/clients/<path:client_ip>/still_working/<path:action_instance_id>', 'client_still_working_on', api_key_required(self.client_still_working_on), methods=['GET'])
        self.flask_app.add_url_rule('/clients/verify_working', 'client_verify_working', api_key_required(self.client_verify_working), methods=['GET'])

        self.flask_app.register_error_handler(HttpException, self.http_exception_handler)
        self.flask_app.register_error_handler(VcsNotFoundException, self.http_exception_handler)

    def http_exception_handler(self, exception):
        response = Response(json.dumps(exception.to_dict()))
        response.status_code = exception.status_code
        response.content_type = 'application/json'
        return response

    @json_response()
    def client_still_working_on(self, client_ip, action_instance_id):
        clients = []
        try:
            clients = StoreService.get_clients(self.flask_app)
        except Exception:
            pass

        if client_ip in clients:
            return {'status': MasterCommunicator.is_still_working_on(action_instance_id, clients[client_ip], self.flask_app.rapid_config.verify_certs)}

        return {"status": "No client found"}

    @json_response()
    def client_verify_working(self):
        action_dal = ActionDal(None, None)
        return action_dal.get_verify_working(self.flask_app.rapid_config.queue_consider_late_time)

    def working_clients(self):
        clients = {}
        for client in MasterCommunicator.get_clients_working_on(StoreService.get_clients(self.flask_app).values(), self.flask_app.rapid_config.verify_certs):
            if client is not None:
                if client['version'] not in clients:
                    clients[client['version']] = []
                clients[client['version']].append(client)
        return Response(json.dumps(clients), content_type='application/json')

    def register_client(self):
        return self.register_request(request)

    def show_clients(self):
        return Response(jsonpickle.encode(self._get_clients()), content_type='application/json')

    def register_request(self, in_request):
        if 'Content-Type' in in_request.headers and in_request.headers['Content-Type'] == 'application/json':
            if 'X-Rapidci-Register-Key' in in_request.headers and in_request.headers['X-Rapidci-Register-Key'] == self.flask_app.rapid_config.register_api_key:
                remote_addr = in_request.remote_addr
                remote_port = in_request.headers['X-Rapidci-port'] if 'X-Rapidci-port' in in_request.headers else None
                time_elapse = max((time.time() * 1000) - float(in_request.headers['X-Rapidci-time']), 1) if 'X-Rapidci-time' in in_request.headers else None

                grains = in_request.json['grains'] if 'grains' in in_request.json else ''
                hostname = in_request.json['hostname'] if 'hostname' in in_request.json else 'unknown'
                grain_restrict = in_request.json['grain_restrict'] if 'grain_restrict' in in_request.json else False

                api_key = in_request.headers['X-Rapidci-Client-Key'] if 'X-Rapidci-Client-Key' in in_request.headers else False
                is_ssl = in_request.headers['X-Is-Ssl'].lower() == 'true' if 'X-is_ssl' in in_request.headers else False

                if remote_port == 443:
                    is_ssl = True

                if not api_key:
                    raise Exception("NO API KEY!")
                client = Client(remote_addr, int(remote_port), grains, grain_restrict, api_key, is_ssl, hostname, time_elapse)

                if HeaderConstants.SINGLE_USE not in in_request.headers:
                    self.store_client(remote_addr, client)

                return Response(jsonpickle.encode(client),
                                content_type='application/json', headers={'Content-Type': 'application/json',
                                                                          'X-Rapidci-Master-Key': self.flask_app.rapid_config.api_key,
                                                                          Version.HEADER: Version.get_version()})
        raise Exception('Not Allowed')

    def store_client(self, remote_addr, definition):
        clients = self._get_clients()
        clients[remote_addr] = definition
        self._save_clients(clients)

    def _save_clients(self, clients):
        StoreService.save_clients(clients, self.flask_app)

    def _get_clients(self):
        return StoreService.get_clients(self.flask_app)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.lib.framework.injectable import Injectable
from rapid.lib.modules import QaModule
from rapid.qa.data.dals.qa_dal import QaDal


class QaService(Injectable, QaModule):
    __injectables__ = {'qa_dal': QaDal}

    def __init__(self, qa_dal):
        """

        :param qa_dal:
        :type qa_dal: QaDal
        """
        self.qa_dal = qa_dal

    def reset_results(self, action_instance_id, session):
        return self.qa_dal.reset_results(action_instance_id, session)

    def save_results(self, action_instance, session, post_data):
        return self.qa_dal.save_results(action_instance, session, post_data)

    def get_test_results(self, action_instance_id, filters=None):
        """
        Get Test Results
        :param action_instance_id:
        :type action_instance_id: int
        :param filters:
        :type filters: string
        :return:
        :rtype:
        """
        return self.qa_dal.get_test_results(action_instance_id, filters=filters)

    def analyze_tests(self, pipeline_instance_id, json):
        """
        Override from QaModule
        :param pipeline_instance_id:
        :type pipeline_instance_id: int
        :param json:
        :type json: dict
        :return:
        :rtype: dict
        """
        return self.qa_dal.analyze_tests(pipeline_instance_id, json)

    def get_qa_testmap_coverage(self, pipeline_instance_id):
        """
        :param pipeline_instance_id:
        :type pipeline_instance_id: int
        :return:
        :rtype:
        """

        return self.qa_dal.get_qa_testmap_coverage(pipeline_instance_id)



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from flask.globals import request
from werkzeug.exceptions import HTTPException

from rapid.qa.qa_service import QaService
from rapid.lib import api_key_required, json_response
from rapid.lib.framework.injectable import Injectable

logger = logging.getLogger("rapid")


class QAController(Injectable):
    __injectables__ = {'qa_service': QaService}

    def __init__(self, qa_service):
        """

        :param qa_service:
        :type qa_service: QaService
        """
        self.app = None
        self.qa_service = qa_service

    def register_url_rules(self, flask_app):
        self.app = flask_app

        flask_app.add_url_rule("/api/qa_tests/analysis", 'qa_tests_analysis', api_key_required(self.analyze_qa_tests), methods=['POST'])
        flask_app.add_url_rule("/api/qa_test_mappings/coverage/<path:pipeline_instance_id>", 'qa_test_mapping_coverage', api_key_required(self.get_qa_testmap_coverage), methods=['GET'])

    @json_response()
    def analyze_qa_tests(self):
        if 'X-Pipeline-Instance-Id' in request.headers:
            try:
                return self.qa_service.analyze_tests(int(request.headers['X-Pipeline-Instance-Id']), request.json)
            except Exception as exception:  # pylint: disable=broad-except
                logger.exception(exception)
                
        raise HTTPException("Missing Pipeline Instance Id")

    @json_response()
    def get_qa_testmap_coverage(self, pipeline_instance_id):
        return self.qa_service.get_qa_testmap_coverage(pipeline_instance_id)



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
# pylint: disable=broad-except
import os
import json
import logging

from rapid.lib.exceptions import InvalidReportException
from rapid.lib.framework.injectable import Injectable
from rapid.master.data.database import execute_db_query

logger = logging.getLogger('rapid')


class ReportDal(Injectable):
    __injectables__ = {'rapid_config': None}

    def __init__(self, rapid_config):
        self._rapid_config = rapid_config
        self._canned_reports = {}
        self._read_report_files()

    def get_canned_report_names(self):
        return list(self._canned_reports.keys())

    def get_canned_report(self, report_name):
        try:
            report = self._canned_reports[report_name]
            results = []
            count = 0
            for row in execute_db_query(report['sql']):
                column_count = 0
                result = {}
                for header in report['headers']:
                    result[header['attribute']] = row[column_count]
                    column_count += 1
                results.append(result)
            return {'headers': report['headers'], 'results': results}
        except Exception:
            pass
        raise InvalidReportException("Report is not found")

    def _read_report_files(self):
        try:
            for root, dirs, files in os.walk(self._rapid_config.custom_reports_dir):  # pylint: disable=unused-variable
                for filename in files:
                    try:
                        basename = os.path.splitext(filename)
                        with open("{}/{}".format(root, filename)) as tmp_file:
                            self._canned_reports[basename[0]] = json.load(tmp_file)
                            self._canned_reports[basename[0]]['sql'] = self._canned_reports[basename[0]]['sql'].replace('%', '%%')
                    except Exception as exception:
                        logger.error("Was Unable to process the directory defined. {}".format(exception))
        except Exception:
            pass



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

try:
    import simplejson as json
except ImportError:
    import json


from werkzeug.exceptions import HTTPException


class HttpException(HTTPException):
    code = 400
    headers = {'Content-Type': 'application/json'}

    def __init__(self, description=None, code=400):
        super(HttpException, self).__init__(description)
        self.code = code

    def get_headers(self, environ=None):
        return self.headers

    def get_body(self, environ=None):
        return json.dumps({"message": self.description})


class ThresholdException(BaseException):
    pass


class ResultsFileNotFoundException(BaseException):
    pass


class ResultsFileNotParsedException(BaseException):
    pass


class UnAuthorizedException(HttpException):
    def __init__(self, message):
        super(UnAuthorizedException, self).__init__(description=message)
        self.code = 401
        self.headers['WWW-Authenticate'] = 'Basic realm="Login Required"'


class VcsNotFoundException(HttpException):
    def __init__(self, message):
        super(VcsNotFoundException, self).__init__(description=message)
        self.code = 401


class DatabaseException(HttpException):
    pass


class InvalidObjectException(HttpException):
    pass


class InvalidReportException(HttpException):
    pass


class ECSLimitReached(Exception):
    pass


class ECSConnectionError(Exception):
    pass


class QueueHandlerShouldSleep(Exception):
    pass



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from rapid.lib.constants import Constants
from rapid.client.parsers.abstract_parser import AbstractParser


class TapParser(AbstractParser):
    @staticmethod
    def get_type():
        return "Tap"

    def _parse_lines(self, lines):
        results = {}
        summary = self.prepare_summary()
        check_stacktrace = False
        current_failure = None
        stacktrace = None

        for line in lines:
            line = line.rstrip()
            if line.startswith("ok"):
                if self.failures_only:
                    continue
                split = line.split(" ", 2)
                results[split[2]] = {'status': Constants.STATUS_SUCCESS}
                summary[Constants.STATUS_SUCCESS] += 1
            elif line.startswith('not ok'):
                check_stacktrace = True
                split = line.split(' ', 4)
                current_failure = split[4]
                stacktrace = ""
            elif line.endswith('...'):
                results[current_failure] = {'status': Constants.STATUS_FAILED, 'stacktrace': stacktrace}
                stacktrace = None
                current_failure = None
                check_stacktrace = False
                summary[Constants.STATUS_FAILED] += 1
            elif check_stacktrace:
                stacktrace += "{}\n".format(line)

        results['__summary__'] = summary
        return results



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import hmac
import logging

from hashlib import sha1

from github import Github
from github.GithubException import GithubException

from rapid.lib.features import Features
from rapid.lib.framework.injectable import Injectable

from rapid.lib.exceptions import HttpException
logger = logging.getLogger("rapid")


class GithubHelper(Injectable):
    __injectables__ = {'rapid_config': 'rapid_config'}
    STATUS_PENDING = "pending"
    STATUS_SUCCESS = "success"
    STATUS_ERROR = "error"
    STATUS_FAILURE = "failure"

    def __init__(self, rapid_config):
        self.user = rapid_config.github_user
        self.password = rapid_config.github_pass

    def get_default_github(self, user=None, password=None):
        if user is None:
            user = self.user
        if password is None:
            password = self.password

        return Github(user, password)

    def record_status(self, request_json, commit_identifier, github=None):
        if Features.is_enabled(Features.TESTING) or not Features.is_enabled(Features.GITHUB_ALERTS):
            return 'Success!'

        if GithubHelper.is_valid_request_json(request_json):
            github = github if github is not None else self.get_default_github()
            repo = github.get_repo(request_json['pull_request_repo'])
            status = request_json['status']
            description = request_json['description']
            context = 'cihub-{}'.format(request_json['type'])
            log_uri = request_json['log_uri'] if 'log_uri' in request_json else ""
            try:
                if log_uri:
                    GithubHelper.set_commit_to_status(repo.get_commit(commit_identifier), status, description, context, target_url=log_uri)
                else:
                    GithubHelper.set_commit_to_status(repo.get_commit(commit_identifier), status, description, context)
                return "Success!"
            except GithubException as exception:
                if 'message' in exception.data and exception.data['message'] == 'Validation Failed':
                    logger.info("LOG_URI:  [{}]".format(log_uri))

                import traceback
                traceback.print_exc()
                raise exception

        raise HttpException("Invalid JSON")

    @staticmethod
    def set_commit_to_status(commit, status, description, context, target_url=None):
        if target_url is not None:
            commit.create_status(status, target_url=target_url, description=description, context=context)
        else:
            commit.create_status(status, description=description, context=context)

    @staticmethod
    def is_valid_request(signature, secret_key, data):
        if not signature or not secret_key or not data:
            return False

        if '=' in signature:
            signature = signature.split('=')[1]
        tmp_sig = GithubHelper.get_signature(secret_key, data)

        return signature == tmp_sig

    @staticmethod
    def get_signature(secret_key, data):
        return hmac.new(secret_key.encode('ascii', 'ignore'), data, sha1).hexdigest()

    @staticmethod
    def is_valid_request_json(request_json):
        assert 'pull_request_repo' in request_json, "Invalid JSON: missing pull_request_repo"
        assert 'status' in request_json, "Invalid JSON: missing status"
        assert 'type' in request_json, "Invalid JSON: missing type"
        assert request_json['status'] in [GithubHelper.STATUS_PENDING, GithubHelper.STATUS_SUCCESS, GithubHelper.STATUS_ERROR, GithubHelper.STATUS_FAILURE], "Invalid JSON: invalid status"

        return True

    @staticmethod
    def get_branch_from_ref(ref):
        return ref[len('refs/heads/'):] if ref is not None else None



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.lib import get_db_session
from rapid.master.data.database.dal.general_dal import GeneralDal
from rapid.workflow.data.models import Status


class StatusDal(GeneralDal):

    def __init__(self, db_session=None):
        self.session = db_session

    def get_status_by_id(self, _id):
        if self.session:
            return self.session.query(Status).get(_id).serialize()
        for session in get_db_session():
            return session.query(Status).get(_id).serialize()



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

initial_schema

Revision ID: 58ebb5ecef90
Revises: 
Create Date: 2015-12-09 13:39:15.566607

"""

# revision identifiers, used by Alembic.
from rapid.workflow.data.models import Status

revision = '58ebb5ecef90'
down_revision = None
branch_labels = ('default',)
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('pipelines',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('statistics',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('statuses',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=True),
    sa.Column('type', sa.String(length=50), nullable=False),
    sa.Column('display_name', sa.String(length=100), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('pipeline_instances',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('created_date', sa.DateTime(), nullable=True),
    sa.Column('start_date', sa.DateTime(), nullable=True),
    sa.Column('end_date', sa.DateTime(), nullable=True),
    sa.Column('pipeline_id', sa.Integer(), nullable=False),
    sa.Column('status_id', sa.Integer(), nullable=False),
    sa.Column('priority', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['pipeline_id'], ['pipelines.id'], ),
    sa.ForeignKeyConstraint(['status_id'], ['statuses.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_pipeline_instances_priority'), 'pipeline_instances', ['priority'], unique=False)
    op.create_table('stages',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=True),
    sa.Column('order', sa.Integer(), nullable=False),
    sa.Column('pipeline_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['pipeline_id'], ['pipelines.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('pipeline_parameters',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('pipeline_instance_id', sa.Integer(), nullable=False),
    sa.Column('parameter', sa.String(length=200), nullable=False),
    sa.Column('value', sa.String(length=4000), nullable=False),
    sa.ForeignKeyConstraint(['pipeline_instance_id'], ['pipeline_instances.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('pipeline_statistics',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('pipeline_instance_id', sa.Integer(), nullable=False),
    sa.Column('statistics_id', sa.Integer(), nullable=False),
    sa.Column('value', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['pipeline_instance_id'], ['pipeline_instances.id'], ),
    sa.ForeignKeyConstraint(['statistics_id'], ['statistics.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('stage_instances',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('created_date', sa.DateTime(), nullable=True),
    sa.Column('start_date', sa.DateTime(), nullable=True),
    sa.Column('end_date', sa.DateTime(), nullable=True),
    sa.Column('order', sa.Integer(), nullable=False),
    sa.Column('status_id', sa.Integer(), nullable=False),
    sa.Column('pipeline_instance_id', sa.Integer(), nullable=False),
    sa.Column('stage_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['pipeline_instance_id'], ['pipeline_instances.id'], ),
    sa.ForeignKeyConstraint(['stage_id'], ['stages.id'], ),
    sa.ForeignKeyConstraint(['status_id'], ['statuses.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('workflows',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=True),
    sa.Column('order', sa.Integer(), nullable=False),
    sa.Column('stage_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['stage_id'], ['stages.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('actions',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('cmd', sa.String(length=255), nullable=False),
    sa.Column('executable', sa.String(length=100), nullable=False),
    sa.Column('args', sa.String(length=255), nullable=False),
    sa.Column('order', sa.Integer(), nullable=False),
    sa.Column('manual', sa.Boolean(), nullable=False),
    sa.Column('grain', sa.String(length=100), nullable=True),
    sa.Column('workflow_id', sa.Integer(), nullable=False),
    sa.Column('pipeline_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['pipeline_id'], ['pipelines.id'], ),
    sa.ForeignKeyConstraint(['workflow_id'], ['workflows.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('workflow_instances',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('created_date', sa.DateTime(), nullable=True),
    sa.Column('start_date', sa.DateTime(), nullable=True),
    sa.Column('end_date', sa.DateTime(), nullable=True),
    sa.Column('order', sa.Integer(), nullable=False),
    sa.Column('status_id', sa.Integer(), nullable=False),
    sa.Column('stage_instance_id', sa.Integer(), nullable=False),
    sa.Column('workflow_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['stage_instance_id'], ['stage_instances.id'], ),
    sa.ForeignKeyConstraint(['status_id'], ['statuses.id'], ),
    sa.ForeignKeyConstraint(['workflow_id'], ['workflows.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_table('action_instances',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('created_date', sa.DateTime(), nullable=True),
    sa.Column('start_date', sa.DateTime(), nullable=True),
    sa.Column('end_date', sa.DateTime(), nullable=True),
    sa.Column('cmd', sa.String(length=255), nullable=False),
    sa.Column('executable', sa.String(length=100), nullable=False),
    sa.Column('args', sa.String(length=255), nullable=False),
    sa.Column('order', sa.Integer(), nullable=False),
    sa.Column('manual', sa.Boolean(), nullable=False),
    sa.Column('slice', sa.String(length=25), nullable=False),
    sa.Column('grain', sa.String(length=100), nullable=True),
    sa.Column('assigned_to', sa.String(length=75), nullable=True),
    sa.Column('status_id', sa.Integer(), nullable=False),
    sa.Column('action_id', sa.Integer(), nullable=False),
    sa.Column('workflow_instance_id', sa.Integer(), nullable=False),
    sa.Column('pipeline_instance_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['pipeline_instance_id'], ['pipeline_instances.id'], ),
    sa.ForeignKeyConstraint(['action_id'], ['actions.id'], ),
    sa.ForeignKeyConstraint(['status_id'], ['statuses.id'], ),
    sa.ForeignKeyConstraint(['workflow_instance_id'], ['workflow_instances.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    ### end Alembic commands ###

    op.bulk_insert(Status.__table__, [
        {"id": 1, "name": "NEW", "display_name": "New", "type": "success", "active": True},
        {"id": 2, "name": "READY", "display_name": "Ready", "type": "success", "active": True},
        {"id": 3, "name": "INPROGRESS", "display_name": "In Progress", "type": "success", "active": True},
        {"id": 4, "name": "SUCCESS", "display_name": "Success", "type": "success", "active": True},
        {"id": 5, "name": "FAILED", "display_name": "Failed", "type": "failed", "active": True},
        {"id": 6, "name": "MERGECONFLICT", "display_name": "Merge Conflict", "type": "failed", "active": True},
        {"id": 7, "name": "UNKOWN", "display_name": "Unknown", "type": "failed", "active": True},
        {"id": 8, "name": "CANCELED", "display_name": "Canceled", "type": "canceled", "active": True},
        {"id": 9, "name": "UNSTABLE", "display_name": "Unstable", "type": "failed", "active": True}
    ])


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('action_instances')
    op.drop_table('workflow_instances')
    op.drop_table('actions')
    op.drop_table('workflows')
    op.drop_table('stage_instances')
    op.drop_table('pipeline_statistics')
    op.drop_table('pipeline_parameters')
    op.drop_table('stages')
    op.drop_index(op.f('ix_pipeline_instances_priority'), table_name='pipeline_instances')
    op.drop_table('pipeline_instances')
    op.drop_table('statuses')
    op.drop_table('statistics')
    op.drop_table('pipelines')
    ### end Alembic commands ###
